# Copyright 2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Keegan Smith

import os
import shutil
from tempfile import TemporaryDirectory, NamedTemporaryFile
import json
import unittest
from unittest.mock import patch

from oaebu_workflows.onix_utils import (
    OnixTransformer,
    collapse_subjects,
    create_personname_fields,
    onix_parser_download,
    onix_parser_execute,
    elevate_product_identifiers,
    normalise_related_products,
    elevate_related_products,
    find_onix_product,
    filter_through_schema,
    remove_invalid_products,
    deduplicate_related_products,
)
from oaebu_workflows.config import test_fixtures_folder, schema_folder
from observatory.platform.observatory_environment import (
    ObservatoryTestCase,
    compare_lists_of_dicts,
)
from observatory.platform.files import load_jsonl

FIXTURES_FOLDER = os.path.join(test_fixtures_folder(), "onix_utils")


class TestOnixTransformer(ObservatoryTestCase):
    """Tests for the ONIX transformer end to end"""

    filtered_name = "filtered.xml"
    errors_removed_name = "errors_removed.xml"
    normalised_name = "normalised.xml"
    deduplicated_name = "deduplicated.xml"
    elevated_name = "elevated.xml"
    parsed_name = "full.jsonl"
    apply_names_name = "name_applied.jsonl"
    collapsed_name = "collapsed.jsonl"

    test_input_metadata = os.path.join(FIXTURES_FOLDER, "input_metadata.xml")
    test_output_parse_only = os.path.join(FIXTURES_FOLDER, "output_parse_only.jsonl")
    test_output_metadata = os.path.join(FIXTURES_FOLDER, "output_metadata.jsonl")

    def test_e2e(self):
        with TemporaryDirectory() as tempdir:
            # Your code here
            transformer = OnixTransformer(
                input_path=self.test_input_metadata,
                output_dir=tempdir,
                filter_products=True,
                error_removal=True,
                normalise_related_products=True,
                deduplicate_related_products=True,
                elevate_related_products=True,
                add_name_fields=True,
                collapse_subjects=True,
                filter_schema=os.path.join(
                    schema_folder(workflow_module="oapen_metadata_telescope"), "oapen_metadata_filter.json"
                ),
                invalid_products_name="invalid_products.xml",
                save_format="jsonl",
                keep_intermediate=True,
            )
            transformer_output_path = transformer.transform()
            self.assert_file_integrity(
                os.path.join(transformer.output_dir, self.filtered_name),
                "7e2b13ae1f25c2d09f11e7864c2f0f92",
                algorithm="md5",
            )
            self.assert_file_integrity(
                os.path.join(transformer.output_dir, self.errors_removed_name),
                "6f6cd81c6abe047fffc95fd7d105d78e",
                algorithm="md5",
            )
            self.assert_file_integrity(
                os.path.join(transformer.output_dir, self.normalised_name),
                "e73e1d4e0eac10d512b343835373be08",
                algorithm="md5",
            )
            self.assert_file_integrity(
                os.path.join(transformer.output_dir, self.deduplicated_name),
                "294a88af4897de5f8b66524afe317520",
                algorithm="md5",
            )
            self.assert_file_integrity(
                os.path.join(transformer.output_dir, self.elevated_name),
                "087b7ebcb806b560acf1d6fcd303817c",
                algorithm="md5",
            )
            self.assert_file_integrity(
                os.path.join(transformer.output_dir, self.parsed_name),
                "033ef3c67663dc042b16be05db2b1a7b",
                algorithm="md5",
            )
            self.assert_file_integrity(
                os.path.join(transformer.output_dir, self.apply_names_name),
                "9dc363f29e8f9bdf8c652dd22aed49e0",
                algorithm="md5",
            )
            self.assert_file_integrity(
                os.path.join(transformer.output_dir, self.collapsed_name),
                "74e4870f5c37a2e9690156b47c3232f7",
                algorithm="md5",
            )
            self.assert_file_integrity(
                os.path.join(transformer.output_dir, "invalid_products.xml"),
                "1ce5155e79ff4e405564038d4520ae3c",
                algorithm="md5",
            )
            compare_lists_of_dicts(
                load_jsonl(self.test_output_metadata),
                load_jsonl(transformer_output_path),
                primary_key="ISBN13",
            )

    def test_save_formats(self):
        """Tests that each of the posible save formats works"""

        def _json_loader(file_path: str):
            with open(file_path, "r") as f:
                return json.load(f)

        for fmt, loader in [("json", _json_loader), ("jsonl", load_jsonl), ("jsonl.gz", load_jsonl)]:
            with TemporaryDirectory() as tempdir:
                transformer = OnixTransformer(
                    input_path=self.test_input_metadata,
                    output_dir=tempdir,
                    save_format=fmt,
                    keep_intermediate=True,
                )
                transformer_output_path = transformer.transform()
                self.assertTrue(os.path.exists(transformer_output_path))
                self.assertTrue(transformer_output_path.endswith(fmt))
                # Try loading the output
                loader(transformer_output_path)


class TestOnixFunctions(ObservatoryTestCase):
    """Tests for the ONIX telescope"""

    def test_onix_parser_download_execute(self):
        """Tests the onix_parser_download and onix_parser_execute functions"""
        with TemporaryDirectory() as tempdir:
            ### Test parser download - fails ###
            with patch("oaebu_workflows.onix_utils.download_file") as mock_download:
                mock_download.return_value = (False, "")
                success, parser_path = onix_parser_download(download_dir=tempdir)
                self.assertEqual(success, False)
                mock_download.return_value = (True, "/this/path/does/not/exist")
                success, parser_path = onix_parser_download(download_dir=tempdir)
                self.assertEqual(success, False)

            ### Test parser download - succeeds ###
            success, parser_path = onix_parser_download(download_dir=tempdir)
            self.assertIn(tempdir, parser_path)
            self.assertEqual(success, True)

            ### Test parser_execute: onix.xml does not exist in input folder ###
            input_dir = os.path.join(tempdir, "input")
            output_dir = os.path.join(tempdir, "output")
            os.mkdir(input_dir)
            os.mkdir(output_dir)
            success = onix_parser_execute(parser_path=parser_path, input_dir=input_dir, output_dir=output_dir)
            self.assertEqual(success, False)

            ### Test parser_execute: nonzero returncode ###
            onix_fixtures = test_fixtures_folder(workflow_module="onix_telescope")
            shutil.copy(os.path.join(onix_fixtures, "20210330_CURTINPRESS_ONIX.xml"), input_dir)
            with patch("oaebu_workflows.onix_utils.wait_for_process") as mock_wfp:
                mock_wfp.return_value = ("stdout", "stderr")
                with patch("oaebu_workflows.onix_utils.subprocess.Popen") as mock_popen:
                    mock_popen.returncode = 1
                    success = onix_parser_execute(parser_path=parser_path, input_dir=input_dir, output_dir=output_dir)
            self.assertEqual(success, False)

            ### Test Case: Successful run ###
            success = onix_parser_execute(parser_path=parser_path, input_dir=input_dir, output_dir=output_dir)
            output_file = os.path.join(output_dir, "full.jsonl")
            self.assertEqual(success, True)
            self.assertTrue(os.path.isfile(output_file))
            self.assert_file_integrity(output_file, "84d46e2942df615f18d270e18e0ebb26", "md5")

    def test_collapse_subjects(self):
        """Tests the thoth_collapse_subjects function"""
        test_subjects_input = os.path.join(FIXTURES_FOLDER, "test_subjects_input.json")
        test_subjects_expected = os.path.join(FIXTURES_FOLDER, "test_subjects_expected.json")
        with open(test_subjects_input, "r") as f:
            onix = json.load(f)
        actual_onix = collapse_subjects(onix)
        with open(test_subjects_expected, "r") as f:
            expected_onix = json.load(f)

        self.assertEqual(len(actual_onix), len(expected_onix))
        self.assertEqual(json.dumps(actual_onix, sort_keys=True), json.dumps(expected_onix, sort_keys=True))


class TestCreatePersonnameFields(unittest.TestCase):
    def test_populate_personname_field(self):
        """Given a list of ONIX products with Contributors that have both NamesBeforeKey and KeyNames fields,
        the function should populate the PersonName field with the concatenation of NamesBeforeKey and KeyNames,
        and also populate the PersonNameInverted field with the concatenation of KeyNames and NamesBeforeKey"""
        input_onix = [
            {
                "Contributors": [
                    {"PersonName": None, "PersonNameInverted": None, "KeyNames": "Doe", "NamesBeforeKey": "John"}
                ]
            },
            {
                "Contributors": [
                    {"PersonName": None, "PersonNameInverted": None, "KeyNames": "Smith", "NamesBeforeKey": "Jane"}
                ]
            },
        ]
        expected_output = [
            {
                "Contributors": [
                    {
                        "PersonName": "John Doe",
                        "PersonNameInverted": "Doe, John",
                        "KeyNames": "Doe",
                        "NamesBeforeKey": "John",
                    }
                ]
            },
            {
                "Contributors": [
                    {
                        "PersonName": "Jane Smith",
                        "PersonNameInverted": "Smith, Jane",
                        "KeyNames": "Smith",
                        "NamesBeforeKey": "Jane",
                    }
                ]
            },
        ]
        output_onix = create_personname_fields(input_onix)
        self.assertEqual(output_onix, expected_output)

    def test_no_modification(self):
        """Should not modify when there is not enough information to populate PersonName or PersonNameInverted"""
        input_onix = [
            {
                "Contributors": [
                    {"PersonName": None, "PersonNameInverted": None, "KeyNames": None, "NamesBeforeKey": "John"}
                ]
            },
            {
                "Contributors": [
                    {"PersonName": None, "PersonNameInverted": None, "KeyNames": "Doe", "NamesBeforeKey": None}
                ]
            },
        ]
        output_onix = create_personname_fields(input_onix)
        self.assertEqual(output_onix, input_onix)  # Should not change input

    def test_prepopulated_personnames(self):
        """Should not modify when PersonName fields are already populated"""
        input_onix = [
            {
                "Contributors": [
                    {
                        "PersonName": "John Doe",
                        "PersonNameInverted": "Doe, John",
                        "KeyNames": "Doe",
                        "NamesBeforeKey": "John",
                    }
                ]
            },
            {
                "Contributors": [
                    {
                        "PersonName": "Wrong Name",
                        "PersonNameInverted": "Doe, Jane",
                        "KeyNames": "Doe",
                        "NamesBeforeKey": "Jane",
                    }
                ]
            },
        ]
        output_onix = create_personname_fields(input_onix)
        self.assertEqual(output_onix, input_onix)  # Should not change input


class TestNormaliseRelatedProducts(unittest.TestCase):
    """Contains tests for normalise_related_products and elevate_product_identifiers"""

    def test_normalise_related_products(self):
        """Should correctly elevate <ProductIdentifier> elements to <RelatedProduct> elements"""
        onix_products = [
            {
                "ProductID": "123",
                "RelatedMaterial": {
                    "RelatedProduct": {
                        "ProductRelationCode": "01",
                        "ProductIdentifier": [
                            {"ProductIDType": "01", "IDValue": "456"},
                            {"ProductIDType": "02", "IDValue": "789"},
                        ],
                    }
                },
            },
            {
                "ProductID": "456",
                "RelatedMaterial": {
                    "RelatedProduct": {
                        "ProductRelationCode": "01",
                        "ProductIdentifier": [
                            {"ProductIDType": "01", "IDValue": "123"},
                            {"ProductIDType": "02", "IDValue": "789"},
                        ],
                    }
                },
            },
            {
                "ProductID": "789",
                "RelatedMaterial": {
                    "RelatedProduct": {
                        "ProductRelationCode": "02",
                        "ProductIdentifier": {"ProductIDType": "03", "IDValue": "ABC"},
                    }
                },
            },
        ]

        expected_result = [
            {
                "ProductID": "123",
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {"ProductRelationCode": "01", "ProductIdentifier": [{"ProductIDType": "01", "IDValue": "456"}]},
                        {"ProductRelationCode": "01", "ProductIdentifier": [{"ProductIDType": "02", "IDValue": "789"}]},
                    ]
                },
            },
            {
                "ProductID": "456",
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {"ProductRelationCode": "01", "ProductIdentifier": [{"ProductIDType": "01", "IDValue": "123"}]},
                        {"ProductRelationCode": "01", "ProductIdentifier": [{"ProductIDType": "02", "IDValue": "789"}]},
                    ]
                },
            },
            {
                "ProductID": "789",
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {"ProductRelationCode": "02", "ProductIdentifier": [{"ProductIDType": "03", "IDValue": "ABC"}]}
                    ]
                },
            },
        ]

        result = normalise_related_products(onix_products)
        self.assertEqual(result, expected_result)

    def test_elevate_product_identifiers(self):
        """Should return a list of <RelatedProduct> structured elements. There should be one elements for each unique
        <ProductIdentifier><IDValue>. If there are multiple <ProductIdentifier><IDValue> elements with different
        <ProductIDType> then these are grouped in the return"""
        related_product = {
            "ProductRelationCode": "ABC",
            "ProductIdentifier": [
                {"ProductIDType": "06", "IDValue": "12345"},
                {"ProductIDType": "15", "IDValue": "12345"},
                {"ProductIDType": "06", "IDValue": "67890"},
            ],
        }
        expected_result = [
            {
                "ProductRelationCode": "ABC",
                "ProductIdentifier": [
                    {"ProductIDType": "06", "IDValue": "12345"},
                    {"ProductIDType": "15", "IDValue": "12345"},
                ],
            },
            {"ProductRelationCode": "ABC", "ProductIdentifier": [{"ProductIDType": "06", "IDValue": "67890"}]},
        ]
        result = elevate_product_identifiers(related_product)
        self.assertCountEqual(result[0]["ProductIdentifier"], expected_result[0]["ProductIdentifier"])
        self.assertCountEqual(result[1]["ProductIdentifier"], expected_result[1]["ProductIdentifier"])

    def test_duplicate_entries(self):
        """Should return a list with a single element when given duplicate entries"""
        related_product = {
            "ProductRelationCode": "ABC",
            "ProductIdentifier": [
                {"ProductIDType": "06", "IDValue": "12345"},
                {"ProductIDType": "06", "IDValue": "12345"},
            ],
        }
        expected_result = [
            {"ProductRelationCode": "ABC", "ProductIdentifier": [{"ProductIDType": "06", "IDValue": "12345"}]}
        ]
        result = elevate_product_identifiers(related_product)
        self.assertEqual(result, expected_result)


class TestElevateRelatedProducts(unittest.TestCase):
    # Should return the same input list if no related products have an ISBN
    def test_no_isbn_related_products(self):
        expected_products = [
            {
                "ProductIdentifier": {"ProductIDType": "15", "IDValue": "1234567890"},
                "RelatedMaterial": {
                    "RelatedProduct": {
                        "ProductRelationCode": "01",
                        "ProductIdentifier": [{"ProductIDType": "03", "IDValue": "9876543210"}],
                    }
                },
            }
        ]

        result = elevate_related_products(expected_products)
        self.assertEqual(result, expected_products)  # Should not change

    # Should return the same input list if no related products are present
    def test_no_related_products(self):
        onix_products = [
            {"ProductIdentifier": {"ProductIDType": "15", "IDValue": "1234567890"}, "RelatedMaterial": {}},
        ]

        result = elevate_related_products(onix_products)
        self.assertEqual(result, onix_products)  # Should not change

    def test_no_parent_isbn(self):
        """Should return the same input list if the parent product has no ISBN"""
        onix_products = [
            {
                "ProductIdentifier": {"ProductIDType": "03", "IDValue": "1234567890"},
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {
                            "ProductRelationCode": "06",
                            "ProductIdentifier": {"ProductIDType": "15", "IDValue": "0987654321"},
                        }
                    ]
                },
            }
        ]

        result = elevate_related_products(onix_products)
        self.assertEqual(result, onix_products)  # Should not change

    def test_relation_code(self):
        """Shuold not elevate related products if the relation code is no '06'"""
        onix_products = [
            {
                "ProductIdentifier": {"ProductIDType": "15", "IDValue": "1234567890"},
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {
                            "ProductRelationCode": "03",
                            "ProductIdentifier": {"ProductIDType": "15", "IDValue": "0987654321"},
                        }
                    ]
                },
            }
        ]

        result = elevate_related_products(onix_products)
        self.assertEqual(result, onix_products)  # Should not change

    def test_duplicate_ids(self):
        """Should not elevate related products if the ID is a duplicate of an existing parent product"""
        onix_products = [
            {
                "ProductIdentifier": {"ProductIDType": "15", "IDValue": "2"},
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {
                            "ProductRelationCode": "06",
                            "ProductIdentifier": {"ProductIDType": "15", "IDValue": "1"},
                        }
                    ]
                },
            },
            {
                "ProductIdentifier": {"ProductIDType": "15", "IDValue": "1"},
                "RelatedMaterial": {"RelatedProduct": []},
            },
        ]

        result = elevate_related_products(onix_products)
        self.assertEqual(result, onix_products)  # Should not change

    def test_related_product_elevation(self):
        """Should elevate ISBN related products to the product level when the parent product has an ISBN"""
        onix_products = [
            {
                "ProductIdentifier": {"ProductIDType": "15", "IDValue": "1"},
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {
                            "ProductRelationCode": "06",
                            "ProductIdentifier": {"ProductIDType": "15", "IDValue": "1.1"},
                        },
                    ]
                },
                "RecordReference": "product1",
            },
            {
                "ProductIdentifier": {"ProductIDType": "15", "IDValue": "2"},
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {
                            "ProductRelationCode": "06",
                            "ProductIdentifier": {"ProductIDType": "15", "IDValue": "2.1"},
                        }
                    ]
                },
                "RecordReference": "product2",
            },
        ]

        expected_result = [
            {
                "ProductIdentifier": {"ProductIDType": "15", "IDValue": "1"},
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {
                            "ProductRelationCode": "06",
                            "ProductIdentifier": {"ProductIDType": "15", "IDValue": "1.1"},
                        },
                    ]
                },
                "RecordReference": "product1",
            },
            {
                "ProductIdentifier": {"ProductIDType": "15", "IDValue": "2"},
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {
                            "ProductRelationCode": "06",
                            "ProductIdentifier": {"ProductIDType": "15", "IDValue": "2.1"},
                        }
                    ]
                },
                "RecordReference": "product2",
            },
            {
                "ProductIdentifier": {"ProductIDType": "15", "IDValue": "1.1"},
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {
                            "ProductRelationCode": "06",
                            "ProductIdentifier": {"ProductIDType": "15", "IDValue": "1"},
                        },
                    ]
                },
                "RecordReference": "product1_1.1",
            },
            {
                "ProductIdentifier": {"ProductIDType": "15", "IDValue": "2.1"},
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {
                            "ProductRelationCode": "06",
                            "ProductIdentifier": {"ProductIDType": "15", "IDValue": "2"},
                        }
                    ]
                },
                "RecordReference": "product2_2.1",
            },
        ]
        result = elevate_related_products(onix_products)
        self.assertEqual(result, expected_result)


class TestFilterThroughSchema(unittest.TestCase):
    def test_filter_through_schema(self):
        """Tests the generic use case of the function"""
        input_dict = {
            # Handle a list of dicts
            "things": [
                {"subthing": {"subsubthing": {"subsubsubthing": "1", "unimportant_subsubsubthing": "1"}}},
                {"subthing": {"subsubthing": {"subsubsubthing": "2"}}, "anotherthing": "2"},
            ],
            "unimportant_thing": "unimportant",  # Neglected field
            "important_thing": "important",  # Field to include
        }
        schema = {
            "things": [{"subthing": {"subsubthing": {"subsubsubthing": None}}, "anotherthing": None}],
            "important_thing": None,
        }
        expected_output = {
            "things": [
                {"subthing": {"subsubthing": {"subsubsubthing": "1"}}},
                {"subthing": {"subsubthing": {"subsubsubthing": "2"}}, "anotherthing": "2"},
            ],
            "important_thing": "important",
        }
        self.assertEqual(filter_through_schema(input_dict, schema), expected_output)

    def test_matching_keys(self):
        """Tests that the function correctly processes the input dictionary when all nested keys match the schema"""
        input_dict = {"thing": {"subthing": {"subsubthing": {"subsubsubthing": "1"}}}}
        schema_dict = {"thing": {"subthing": {"subsubthing": {"subsubsubthing": None}}}}
        expected_output = {"thing": {"subthing": {"subsubthing": {"subsubsubthing": "1"}}}}
        self.assertEqual(filter_through_schema(input_dict, schema_dict), expected_output)

    def test_empty_input(self):
        """Tests that the function correctly handles an empty input dictionary"""
        input_dict = {}
        schema_dict = {"thing": {"subthing": {"subsubthing": None}}}
        expected_output = {}
        self.assertEqual(filter_through_schema(input_dict, schema_dict), expected_output)

    def test_no_matching_keys(self):
        """Tests that the function correctly handles an input dictionary with no matching keys in the schema"""
        input_dict = {"thing": {"subthing": {"subsubthing": "1"}}}
        schema_dict = {"other_thing": {"other_subthing": {"other_subsubthing": None}}}
        expected_output = {}
        self.assertEqual(filter_through_schema(input_dict, schema_dict), expected_output)

    def test_edge_case_empty_schema(self):
        """Tests that the function correctly handles an empty schema"""
        input_dict = {"thing": {"subthing": {"subsubthing": "1"}}}
        schema_dict = {}
        expected_output = {}
        self.assertEqual(filter_through_schema(input_dict, schema_dict), expected_output)


class TestRemoveInvalidProducts(unittest.TestCase):
    oapen_metadata_fixtures = test_fixtures_folder(workflow_module="oapen_metadata_telescope")
    valid_parsed_xml = os.path.join(oapen_metadata_fixtures, "parsed_valid.xml")
    invalid_products_removed_xml = os.path.join(oapen_metadata_fixtures, "invalid_products_removed.xml")
    empty_xml = os.path.join(oapen_metadata_fixtures, "empty_download.xml")
    invalid_products_xml = os.path.join(oapen_metadata_fixtures, "invalid_products.xml")

    def test_remove_invalid_products(self):
        """Tests the function used to remove invalid products from an xml file"""
        with NamedTemporaryFile() as invalid_products_file, NamedTemporaryFile() as processed_file:
            remove_invalid_products(
                self.valid_parsed_xml, processed_file.name, invalid_products_file=invalid_products_file.name
            )

            # Make assertions of the processed xml
            with open(processed_file.name) as f:
                processed_xml = f.readlines()
            with open(self.invalid_products_removed_xml) as f:
                assertion_xml = f.readlines()
            self.assertEqual(len(processed_xml), len(assertion_xml))
            self.assertEqual(processed_xml, assertion_xml)

            # Make assertions of the invalid products
            with open(invalid_products_file.name) as f:
                invalid_products_xml = f.readlines()
            with open(self.invalid_products_xml) as f:
                assertion_xml = f.readlines()
            self.assertEqual(len(invalid_products_xml), len(assertion_xml))
            self.assertEqual(invalid_products_xml, assertion_xml)

    def test_empty_xml(self):
        """Tests the function used to remove invalid products from an xml file"""
        with NamedTemporaryFile() as invalid_products_file, NamedTemporaryFile() as processed_file:
            self.assertRaises(
                AttributeError,
                remove_invalid_products,
                self.empty_xml,
                processed_file.name,
                invalid_products_file=invalid_products_file.name,
            )


class TestFindOnixProduct(unittest.TestCase):
    # Tests that the function can extract multiple products from a valid input xml
    valid_input = [
        "<ONIXMessage>",
        "<Product>",
        "<RecordReference>1</RecordReference>",
        "</Product>",
        "<Product>",
        "<RecordReference>2</RecordReference>",
        "</Product>",
        "<Product>",
        "<RecordReference>3</RecordReference>",
        "<SomeOtherField>something else</SomeOtherField>",
        "</Product>",
        "</ONIXMessage>",
    ]

    def test_find_onix_product(self):
        """Test that the function can extract multiple products from a valid input xml"""
        output = find_onix_product(self.valid_input, 1)
        self.assertEqual(output.record_reference, "1")
        self.assertEqual(output.product, {"RecordReference": "1"})
        output = find_onix_product(self.valid_input, 8)
        self.assertEqual(output.record_reference, "3")
        self.assertEqual(output.product, {"RecordReference": "3", "SomeOtherField": "something else"})

    def test_out_of_bounds_supplied(self):
        """Test that errors are thrown when improper input is supplied"""
        self.assertRaisesRegex(IndexError, "not within the length of the file", find_onix_product, self.valid_input, -1)
        self.assertRaisesRegex(
            IndexError, "not within the length of the file", find_onix_product, self.valid_input, len(self.valid_input)
        )

    def test_missing_record_reference(self):
        """Tests that a product without a RecordReference raises a KeyError"""
        all_lines = ["<Product>", "<RandomField>1</RandomField>", "</Product>"]
        self.assertRaises(KeyError, find_onix_product, all_lines, 1)

    def test_empty_product(self):
        """Tests that a product without a RecordReference raises a KeyError"""
        all_lines = ["<Product>", "</Product>"]
        self.assertRaisesRegex(ValueError, "Product field is empty", find_onix_product, all_lines, 1)

    def test_no_product_tags(self):
        """Tests that the function raises a ValueError when <Product> tags are not closed or missing"""
        missing_product_xml = ["<ONIXMessage>", "</ONIXMessage>"]
        with self.assertRaisesRegex(ValueError, "Product not found surrounding line"):
            find_onix_product(missing_product_xml, 0)
        missing_product_xml = ["<ONIXMessage>", "<Product>", "</ONIXMessage>"]
        with self.assertRaisesRegex(ValueError, "Product not found surrounding line"):
            find_onix_product(missing_product_xml, 1)
        missing_product_xml = ["<ONIXMessage>", "</Product>" "</ONIXMessage>"]
        with self.assertRaisesRegex(ValueError, "Product not found surrounding line"):
            find_onix_product(missing_product_xml, 1)


class TestDeduplicateRelatedProducts(unittest.TestCase):
    def test_no_duplicated_related_products(self):
        """Should return the same list if there are no duplicated related products in any of the products"""
        onix_products = [
            {
                "ProductIdentifier": [{"ProductIDType": "03", "IDValue": "1234567890"}],
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {
                            "ProductRelationCode": "01",
                            "ProductIdentifier": [{"ProductIDType": "03", "IDValue": "0987654321"}],
                        },
                        {
                            "ProductRelationCode": "02",
                            "ProductIdentifier": [{"ProductIDType": "03", "IDValue": "1357924680"}],
                        },
                    ]
                },
            },
            {
                "ProductIdentifier": [{"ProductIDType": "03", "IDValue": "2468135790"}],
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {
                            "ProductRelationCode": "03",
                            "ProductIdentifier": [{"ProductIDType": "03", "IDValue": "9876543210"}],
                        },
                        {
                            "ProductRelationCode": "04",
                            "ProductIdentifier": [{"ProductIDType": "03", "IDValue": "0246813579"}],
                        },
                    ]
                },
            },
        ]
        result = deduplicate_related_products(onix_products)
        self.assertEqual(result, onix_products)

    def test_remove_duplicated_related_products_fixed(self):
        """Should remove duplicated related products from a single product"""
        onix_products = [
            {
                "ProductIdentifier": [{"ProductIDType": "03", "IDValue": "1234567890"}],
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {
                            "ProductRelationCode": "01",
                            "ProductIdentifier": [{"ProductIDType": "03", "IDValue": "0987654321"}],
                        },
                        {
                            "ProductRelationCode": "02",
                            "ProductIdentifier": [{"ProductIDType": "03", "IDValue": "0987654321"}],
                        },
                    ]
                },
            }
        ]
        expected_result = [
            {
                "ProductIdentifier": [{"ProductIDType": "03", "IDValue": "1234567890"}],
                "RelatedMaterial": {
                    "RelatedProduct": [
                        {
                            "ProductRelationCode": "02",
                            "ProductIdentifier": [{"ProductIDType": "03", "IDValue": "0987654321"}],
                        }
                    ]
                },
            }
        ]
        result = deduplicate_related_products(onix_products)
        self.assertEqual(result, expected_result)

    def test_empty_input_list(self):
        """Should return an empty list if the input list is empty"""
        onix_products = []
        result = deduplicate_related_products(onix_products)
        self.assertEqual(result, [])



# Copyright 2020-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, Keegan Smith

import os
from unittest import TestCase
from unittest.mock import patch, call

import pendulum
from airflow.utils.state import State
from airflow.models.connection import Connection
import vcr

from oaebu_workflows.config import test_fixtures_folder
from oaebu_workflows.oaebu_partners import partner_from_str
from oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope import (
    UclDiscoveryTelescope,
    get_isbn_eprint_mappings,
    download_discovery_stats,
    transform_discovery_stats,
)
from observatory.platform.api import get_dataset_releases
from observatory.platform.observatory_config import Workflow
from observatory.platform.bigquery import bq_table_id
from observatory.platform.gcs import gcs_blob_name_from_path
from observatory.platform.observatory_environment import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    find_free_port,
    load_and_parse_json,
)


class TestUclDiscoveryTelescope(ObservatoryTestCase):
    """Tests for the Ucl Discovery telescope"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests."""
        super(TestUclDiscoveryTelescope, self).__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")
        fixtures_folder = test_fixtures_folder(workflow_module="ucl_discovery_telescope")
        self.download_cassette = os.path.join(fixtures_folder, "download_cassette.yaml")
        self.test_table = os.path.join(fixtures_folder, "test_table.json")

    def test_dag_structure(self):
        """Test that the UCL Discovery DAG has the correct structure."""

        dag = UclDiscoveryTelescope(
            dag_id="Test_Dag", cloud_workspace=self.fake_cloud_workspace, sheet_id="foo"
        ).make_dag()
        self.assert_dag_structure(
            {
                "check_dependencies": ["download"],
                "download": ["upload_downloaded"],
                "upload_downloaded": ["transform"],
                "transform": ["upload_transformed"],
                "upload_transformed": ["bq_load"],
                "bq_load": ["add_new_dataset_releases"],
                "add_new_dataset_releases": ["cleanup"],
                "cleanup": [],
            },
            dag,
        )

    def test_dag_load(self):
        """Test that the UCL Discovery DAG can be loaded from a DAG bag."""
        env = ObservatoryEnvironment(
            workflows=[
                Workflow(
                    dag_id="ucl_discovery",
                    name="UCL Discovery Telescope",
                    class_name="oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.UclDiscoveryTelescope",
                    cloud_workspace=self.fake_cloud_workspace,
                    kwargs=dict(sheet_id="foo"),
                )
            ]
        )
        with env.create():
            self.assert_dag_load_from_config("ucl_discovery")

    def test_telescope(self):
        """Test the UCL Discovery telescope end to end."""
        # Setup Observatory environment
        env = ObservatoryEnvironment(
            self.project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )

        # Setup Telescope
        data_partner = partner_from_str("ucl_discovery")
        data_partner.bq_dataset_id = env.add_dataset()
        telescope = UclDiscoveryTelescope(
            dag_id="ucl_discovery",
            cloud_workspace=env.cloud_workspace,
            sheet_id="foo",
            data_partner=data_partner,
            max_threads=1,
        )
        dag = telescope.make_dag()
        execution_date = pendulum.datetime(year=2023, month=6, day=1)

        # Create the Observatory environment and run tests
        with env.create(), env.create_dag_run(dag, execution_date):
            # env.add_connection(Connection(conn_id=telescope.oaebu_service_account_conn_id))
            # Mock return values of download function
            interval_start = pendulum.instance(env.dag_run.data_interval_start)
            sheet_return = [
                ["ISBN13", "discovery_eprintid", "date", "title_list_title"],
                ["ISBN_1", "eprint_id1", interval_start.add(days=10).format("YYYYMMDD"), "title1"],
                ["ISBN_2", "", interval_start.add(days=10).format("YYYYMMDD"), "title2"],  # should be ignored
                ["ISBN_3", "eprint_id3", interval_start.add(years=1).format("YYYYMMDD"), "title3"],  # should be ignored
                ["", "eprint_id4", interval_start.add(days=10).format("YYYYMMDD"), "title4"],  # should be ignored
                ["ISBN_5", "eprint_id5", interval_start.subtract(months=5).format("YYYYMMDD"), "title5"],
            ]
            conn = Connection(
                conn_id="oaebu_service_account",
                uri=f"google-cloud-platform://?type=service_account&private_key_id=private_key_id"
                f"&private_key=private_key"
                f"&client_email=client_email"
                f"&client_id=client_id"
                f"&token_uri=token_uri",
            )
            env.add_connection(conn)

            ############################
            ### Main telescope tasks ###
            ############################

            # Test that all dependencies are specified: no error should be thrown
            ti = env.run_task(telescope.check_dependencies.__name__)

            # download
            cassette = vcr.VCR(record_mode="none")
            sa_patch = patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.service_account")
            conn_patch = patch(
                "oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.BaseHook.get_connection"
            )
            build_patch = patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.discovery.build")
            with sa_patch, conn_patch, build_patch as mock_build, cassette.use_cassette(self.download_cassette):
                mock_service = mock_build.return_value.spreadsheets.return_value.values.return_value.get.return_value
                mock_service.execute.return_value = {"values": sheet_return}
                ti = env.run_task(telescope.download.__name__)
            self.assertEqual(ti.state, State.SUCCESS)

            # upload_downloaded
            ti = env.run_task(telescope.upload_downloaded.__name__)
            self.assertEqual(ti.state, State.SUCCESS)

            # transform
            with sa_patch, conn_patch, build_patch as mock_build:
                mock_service = mock_build.return_value.spreadsheets.return_value.values.return_value.get.return_value
                mock_service.execute.return_value = {"values": sheet_return}
                ti = env.run_task(telescope.transform.__name__)
            self.assertEqual(ti.state, State.SUCCESS)

            # upload_transformed
            ti = env.run_task(telescope.upload_transformed.__name__)
            self.assertEqual(ti.state, State.SUCCESS)

            # bq_load
            ti = env.run_task(telescope.bq_load.__name__)
            self.assertEqual(ti.state, State.SUCCESS)

            ##############################################
            ### Create the release and make assertions ###
            ##############################################

            release = telescope.make_release(
                run_id=env.dag_run.run_id,
                data_interval_start=pendulum.parse(str(env.dag_run.data_interval_start)),
                data_interval_end=pendulum.parse(str(env.dag_run.data_interval_end)),
            )

            # Download
            self.assertTrue(os.path.exists(release.download_country_path))
            self.assertTrue(os.path.exists(release.download_totals_path))

            # Upload Downloaded
            download_country_blob = gcs_blob_name_from_path(release.download_country_path)
            self.assert_blob_integrity(env.download_bucket, download_country_blob, release.download_country_path)
            download_totals_blob = gcs_blob_name_from_path(release.download_totals_path)
            self.assert_blob_integrity(env.download_bucket, download_totals_blob, release.download_totals_path)

            # Transform
            self.assertTrue(os.path.exists(release.transform_path))

            # Upload Transform
            self.assert_blob_integrity(
                env.transform_bucket, gcs_blob_name_from_path(release.transform_path), release.transform_path
            )

            # Bigquery load
            table_id = bq_table_id(
                telescope.cloud_workspace.project_id,
                telescope.data_partner.bq_dataset_id,
                telescope.data_partner.bq_table_name,
            )
            self.assert_table_integrity(table_id, 2)
            self.assert_table_content(
                table_id, load_and_parse_json(self.test_table, date_fields="release_date"), "ISBN"
            )

            ###################
            ### Final tasks ###
            ###################

            # Add_dataset_release_task
            dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
            self.assertEqual(len(dataset_releases), 0)
            ti = env.run_task(telescope.add_new_dataset_releases.__name__)
            self.assertEqual(ti.state, State.SUCCESS)
            dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
            self.assertEqual(len(dataset_releases), 1)

            # Test that all telescope data deleted
            ti = env.run_task(telescope.cleanup.__name__)
            self.assertEqual(ti.state, State.SUCCESS)
            self.assert_cleanup(release.workflow_folder)


class TestGetIsbnEprintMappings(TestCase):
    """Tests for the get_isbn_eprint_mappings function"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # Set the cutoff date for the tests
        self.cutoff_date = pendulum.datetime(year=2023, month=6, day=30)

    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.service_account")
    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.BaseHook.get_connection")
    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.discovery.build")
    def test_get_isbn_eprint_mappings(self, mock_build, mock_get_connection, mock_sa):
        # Mock the Google Sheets API response
        sheet_contents = [
            ["ISBN13", "discovery_eprintid", "date", "title_list_title"],
            ["111", "eprint_1", "2023-08-01", "title1"],  # past cutoff, should be ignored
            ["222", "eprint_2", "2023-06-01", "title2"],
            ["333", "eprint_3", "2023-07-01", "title3"],  # past cutoff, should be ignored
            ["444", "eprint_4", "2023-06-30", "title4"],
        ]
        mock_service = mock_build.return_value.spreadsheets.return_value.values.return_value.get.return_value
        mock_service.execute.return_value = {"values": sheet_contents}

        # Call the function to test
        mappings = get_isbn_eprint_mappings("sheet_id", "service_account_conn_id", self.cutoff_date)

        # Assert that the returned mappings match the expected mappings
        expected_mappings = {
            "eprint_2": {"ISBN13": "222", "title": "title2"},
            "eprint_4": {"ISBN13": "444", "title": "title4"},
        }
        self.assertEqual(mappings, expected_mappings)

    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.service_account")
    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.BaseHook.get_connection")
    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.discovery.build")
    def test_invalid_header(self, mock_build, mock_get_connection, mock_sa):
        # Mock the Google Sheets API response with an invalid header
        invalid_sheet_contents = [
            [
                "ISBN13",
                "discovery_id",
                "date",
                "title_list_title",
            ],  # Invalid header, should contain "discovery_eprintid"
            ["222", "eprint_2", "2023-06-01", "title2"],
        ]
        mock_service = mock_build.return_value.spreadsheets.return_value.values.return_value.get.return_value
        mock_service.execute.return_value = {"values": invalid_sheet_contents}

        with self.assertRaisesRegex(ValueError, "Invalid header found"):
            get_isbn_eprint_mappings("sheet_id", "service_account_conn_id", self.cutoff_date)

    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.service_account")
    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.BaseHook.get_connection")
    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.discovery.build")
    def test_empty_sheet(self, mock_build, mock_get_connection, mock_sa):
        # Mock the Google Sheets API response with an empty sheet
        mock_build.return_value.spreadsheets.return_value.values.return_value.get.return_value.execute.return_value = {}

        with self.assertRaisesRegex(ValueError, "No content found"):
            get_isbn_eprint_mappings("sheet_id", "service_account_conn_id", self.cutoff_date)

    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.service_account")
    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.BaseHook.get_connection")
    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.discovery.build")
    def test_missing_values(self, mock_build, mock_get_connection, mock_sa):
        # Mock the Google Sheets API response with a missing value
        sheet_contents = [
            ["ISBN13", "discovery_eprintid", "date", "title_list_title"],
            ["111", "", "2023-06-01", "title1"],  # eprint ID missing
            ["", "eprint_2", "2023-06-01", "title2"],  # ISBN missing
            ["333", "eprint_3", "2023-06-01", ""],  # Title missing, should still pass
        ]
        mock_service = mock_build.return_value.spreadsheets.return_value.values.return_value.get.return_value
        mock_service.execute.return_value = {"values": sheet_contents}

        mappings = get_isbn_eprint_mappings("sheet_id", "service_account_conn_id", self.cutoff_date)
        expected_mappings = {"eprint_3": {"ISBN13": "333", "title": ""}}
        self.assertEqual(mappings, expected_mappings)


class TestDownloadDiscoveryStats(TestCase):
    """Tests for the download_discovery_stats function"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # Set the cutoff date for the tests
        self.start_date = pendulum.datetime(2022, 1, 1)
        self.end_date = pendulum.datetime(2022, 1, 31)
        self.start_formatted = self.start_date.format("YYYYMMDD")
        self.end_formatted = self.end_date.format("YYYYMMDD")
        self.eprint_id = "12345"

    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.retry_get_url")
    def test_download_discovery_stats(self, mock_retry_get_url):
        """Test the download_discovery_stats function works with correct inputs"""
        expected_countries_url = (
            "https://discovery.ucl.ac.uk/cgi/stats/get"
            f"?from={self.start_date.format('YYYYMMDD')}&to={self.end_date.format('YYYYMMDD')}"
            f"&irs2report=eprint&set_name=eprint&set_value={self.eprint_id}&datatype=countries&top=countries"
            "&view=Table&limit=all&export=JSON"
        )
        expected_totals_url = (
            "https://discovery.ucl.ac.uk/cgi/stats/get"
            f"?from={self.start_date.format('YYYYMMDD')}&to={self.end_date.format('YYYYMMDD')}"
            f"&irs2report=eprint&set_name=eprint&set_value={self.eprint_id}&datatype=downloads&graph_type=column"
            "&view=Google%3A%3AGraph&date_resolution=month&title=Download+activity+-+last+12+months&export=JSON"
        )
        http_returns = [
            {"timescale": {"from": self.start_formatted, "to": self.end_formatted}, "set": {"value": self.eprint_id}},
            {"timescale": {"from": self.start_formatted, "to": self.end_formatted}, "set": {"value": self.eprint_id}},
        ]
        mock_retry_get_url.return_value.json.side_effect = http_returns

        # Check that the correct data is returned
        result = download_discovery_stats(self.eprint_id, self.start_date, self.end_date)

        # Check that constructed urls are correct
        expected_calls = [call(expected_countries_url), call().json(), call(expected_totals_url), call().json()]
        mock_retry_get_url.assert_has_calls(expected_calls)

        # Check that returned results are correct
        self.assertEqual(result[0], http_returns[0])
        self.assertEqual(result[1], http_returns[1])

    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.retry_get_url")
    def test_download_discovery_stats_invalid_timescale(self, mock_retry_get_url):
        """Check if exceptions raised when timescale is inconsistent with inputs"""
        mock_retry_get_url.return_value.json.side_effect = [
            {"timescale": {"from": "19700101", "to": "19700130"}, "set": {"value": self.eprint_id}},
            {"timescale": {"from": "19700101", "to": "19700130"}, "set": {"value": self.eprint_id}},
        ]
        self.assertRaisesRegex(
            ValueError, "timescale", download_discovery_stats, self.eprint_id, self.start_date, self.end_date
        )

    @patch("oaebu_workflows.ucl_discovery_telescope.ucl_discovery_telescope.retry_get_url")
    def test_download_discovery_stats_invalid_eprint_id(self, mock_retry_get_url):
        """Check if exceptions raised when eprint ID is inconsistent with inputs"""
        mock_retry_get_url.return_value.json.side_effect = [
            {"timescale": {"from": self.start_formatted, "to": self.end_formatted}, "set": {"value": "67890"}},
            {"timescale": {"from": self.start_formatted, "to": self.end_formatted}, "set": {"value": "67890"}},
        ]
        self.assertRaisesRegex(
            ValueError, "eprint ID", download_discovery_stats, self.eprint_id, self.start_date, self.end_date
        )


class TestTransformDiscoveryStats(TestCase):
    def test_transform_discovery_stats(self):
        """Test the transform_discovery_stats function when inputs are valid"""
        country_record = {
            "set": {"value": "eprint_id_1"},
            "timescale": {"to": "20230630", "format": "YYYYMMDD", "from": "20230601"},
            "origin": "foo",
            "records": [{"value": "AA", "count": "42"}, {"value": "BB", "count": "42"}],
        }
        totals_record = {
            "set": {"value": "eprint_id_1"},
            "timescale": {"to": "20230630", "format": "YYYYMMDD", "from": "20230601"},
            "origin": "foo",
            "records": [{"count": "84", "datestamp": "202306", "description": "foo"}],
        }
        isbn = "ISBN_1"
        title = "title1"
        expected_output = {
            "ISBN": "ISBN_1",
            "title": "title1",
            "eprint_id": "eprint_id_1",
            "timescale": {"to": "20230630", "format": "YYYYMMDD", "from": "20230601"},
            "origin": "foo",
            "total_downloads": "84",
            "country": [{"value": "AA", "count": "42"}, {"value": "BB", "count": "42"}],
        }
        self.assertEqual(transform_discovery_stats(country_record, totals_record, isbn, title), expected_output)

    def test_transform_discovery_stats_no_country_records(self):
        """Test the transform_discovery_stats function when country records are missing"""
        country_record = {
            "set": {"value": "eprint_id_1"},
            "timescale": {"to": "20230630", "format": "YYYYMMDD", "from": "20230601"},
            "origin": "foo",
            # Missing records
        }
        totals_record = {
            "set": {"value": "eprint_id_1"},
            "timescale": {"to": "20230630", "format": "YYYYMMDD", "from": "20230601"},
            "origin": "foo",
            "records": [{"count": "84", "datestamp": "202306", "description": "foo"}],
        }
        isbn = "ISBN_1"
        title = "title1"
        expected_output = {
            "ISBN": "ISBN_1",
            "title": "title1",
            "eprint_id": "eprint_id_1",
            "timescale": {"to": "20230630", "format": "YYYYMMDD", "from": "20230601"},
            "origin": "foo",
            "total_downloads": "84",
            "country": [],
        }
        self.assertEqual(transform_discovery_stats(country_record, totals_record, isbn, title), expected_output)

    def test_transform_discovery_stats_mismatching_eprint_ids(self):
        """Test the transform_discovery_stats function when eprint IDs do not match"""
        country_record = {
            "set": {"value": "eprint_id_1"},
            "timescale": {"to": "20230630", "format": "YYYYMMDD", "from": "20230601"},
            "origin": "foo",
            "records": [{"value": "AA", "count": "42"}, {"value": "BB", "count": "42"}],
        }
        totals_record = {
            "set": {"value": "eprint_id_2"},
            "timescale": {"to": "20230630", "format": "YYYYMMDD", "from": "20230601"},
            "origin": "foo",
            "records": [{"count": "84", "datestamp": "202306", "description": "foo"}],
        }
        isbn = "ISBN_3"
        title = "title3"
        with self.assertRaisesRegex(ValueError, "eprint ID do not match"):
            transform_discovery_stats(country_record, totals_record, isbn, title)

    def test_transform_discovery_stats_mismatching_timescales(self):
        """Test the transform_discovery_stats function when timescales do not match"""
        country_record = {
            "set": {"value": "eprint_id_1"},
            "timescale": {"to": "19700131", "format": "YYYYMMDD", "from": "19700101"},
            "origin": "foo",
            "records": [{"value": "AA", "count": "42"}, {"value": "BB", "count": "42"}],
        }
        totals_record = {
            "set": {"value": "eprint_id_1"},
            "timescale": {"to": "20230630", "format": "YYYYMMDD", "from": "20230601"},
            "origin": "foo",
            "records": [{"count": "84", "datestamp": "202306", "description": "foo"}],
        }
        isbn = "ISBN_3"
        title = "title3"
        with self.assertRaisesRegex(ValueError, "Timescales do not match"):
            transform_discovery_stats(country_record, totals_record, isbn, title)



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Tuan Chien


from typing import Dict, List, Set, Tuple, Union

import numpy as np


class UnionFind:
    """
    Union Find using weighted quick union and path compression.
    Instead of working on objects and requiring further implementation of comparators and size methods,
    this will operate on integers. Users should handle the mapping from objects to a distinct integer, e.g., by using
    the index of the objects if they are in a list.
    See: https://en.wikipedia.org/wiki/Disjoint-set_data_structure
    and https://www.cs.princeton.edu/~rs/AlgsDS07/01UnionFind.pdf
    """

    def __init__(self, size: int):
        """
        :param size: Number of elements we are dealing with in total.
        """

        self.id = np.arange(size)  # Root mapping.
        self.tree_sizes = [1] * size

    def root(self, node: int) -> int:
        """
        Find the root (representative) of an object.  Update root mappings along the way (path compression).
        :param node: Object to find the root for.
        :return: The object's root representative.
        """

        while node != self.id[node]:
            self.id[node] = self.id[self.id[node]]
            node = self.id[node]

        return node

    def find(self, p: int, q: int) -> bool:
        """
        Check if two objects have the same root representative.
        :param p: First object.
        :param q: Second object.
        :return: Whether p, q have the same root representative.
        """

        return self.root(p) == self.root(q)

    def unite(self, p: int, q: int):
        """
        Merge two objects into the same class, i.e., make the two objects have the same root representative.
        Use weighted union to merge smaller trees into the bigger trees.
        :param p First object.
        :param q Second object.
        """

        i = self.root(p)
        j = self.root(q)

        if self.find(p, q):
            return

        if self.tree_sizes[i] < self.tree_sizes[j]:
            self.id[i] = j
            self.tree_sizes[j] + self.tree_sizes[i]
        else:
            self.id[j] = i
            self.tree_sizes[i] += self.tree_sizes[j]

    def get_partition(self) -> List[List[int]]:
        """
        Get the current class partition of the objects.
        :return: Partition of the objects as a list of lists (no guaranteed ordering).
        """

        n = len(self.id)

        partition = {}
        for i in range(n):
            root = self.root(i)
            if root not in partition:
                partition[root] = [i]
            else:
                partition[root].append(i)

        return list(partition.values())


class BookWork:
    """
    A book work is an abstract entity comprising the intellectual property embodied in a manifestation.
    Works manifest themselves as products of different form, e.g., paperback, PDF.
    """

    def __init__(self, *, work_id: str, work_id_type: str, products: List[Dict]):
        """
        :param work_id: The Work ID.
        :param work_id_type: Type scheme used in the Work ID.
        :param products: List of products that manifest the work.
        """

        self.work_id = work_id
        self.work_id_type = work_id_type
        self.products = products
        self.isbns = set([product["ISBN13"] for product in products])

    def add_product(self, product: dict):
        """
        Add a product to the work.
        :param product: A product record.
        """

        isbn = product["ISBN13"]
        if isbn in self.isbns:
            return

        self.products.append(product)
        self.isbns.add(isbn)


class BookWorkFamily:
    """
    A book work family aggregates different editions of a work together.
    """

    def __init__(self, *, works: List[BookWork], work_family_id=None, work_family_id_type: Union[None, str] = None):
        """
        :param works: List of works in the family.
        :param work_family_id: Work family ID.
        :param work_family_id_type: Type of Work family ID.
        """

        self.works = works
        self.work_family_id = work_family_id
        self.work_family_id_type = work_family_id_type
        self.isbns = set()


def get_pref_product_id(relprod: dict) -> Tuple[str, str]:
    """
    Get the most preferred product ID. It will return the one with the lowest preference
    number in the id_pref list, or an arbitrary identifier from the list if the listed preferences are not found.
    :param relprod: Related product record.
    :return: The Product ID type, and Product ID as a pair.
    """

    if "ISBN13" in relprod and relprod["ISBN13"] is not None:
        return "ISBN13", relprod["ISBN13"]

    if "GTIN_13" in relprod and relprod["GTIN_13"] is not None:
        return "GTIN_13", relprod["GTIN_13"]

    if "DOI" in relprod and relprod["DOI"] is not None:
        return "DOI", relprod["DOI"]

    if "PID_Proprietary" in relprod and relprod["PID_Proprietary"] is not None:
        return "PID_Proprietary", relprod["PID_Proprietary"]

    raise Exception(f"Unhandled product identifier found in {relprod}")


class BookWorkAggregator:
    """
    Aggregates ONIX records into "works" (BookWork object). If the WorkID exists in the ONIX record, it will use that.
    The order of preference for the work identifier types is: ISBN-13 > DOI > Proprietary > everything else.
    If no identifier is specified in ONIX, i.e., no RelatedWorks info, one of the ISBN13 in the work will be used as a
    representative ID.
    """

    def __init__(self, records: List[dict]):
        """
        :param records: List of ONIX Product records.
        """

        self.relevant_product_codes = self.set_relevant_product_relation_codes()
        self.errors = list()
        self.works = list()

        self.records = self.filter_out_duplicate_records(records)

        self.n = len(self.records)
        self.isbn13_to_index = {self.records[i]["ISBN13"]: i for i in range(self.n)}

        self.gtin13_to_product = {}
        for record in self.records:
            if "GTIN_13" in record and record["GTIN_13"] is not None:
                gtin = record["GTIN_13"]
                self.gtin13_to_product[gtin] = record

        self.proprietary_to_product = {}
        for record in self.records:
            if "PID_Proprietary" in record and record["PID_Proprietary"] is not None:
                pid_proprietary = record["PID_Proprietary"]
                self.proprietary_to_product[pid_proprietary] = record

        self.uf = UnionFind(self.n)

    def filter_out_duplicate_records(self, records: dict) -> List[dict]:
        """Filter out records with duplicate ISBNs.  Logs the duplicates, and returns the filtered records.

        :param records: Product records.
        :return: Tuple of a list of filtered records, and a list of ISBNs which appear more than once in a record.
        """

        isbns = set()
        filtered_records = list()
        duplicates = list()

        for record in records:
            isbn = record["ISBN13"]
            if isbn not in isbns:
                isbns.add(isbn)
                filtered_records.append(record)
            else:
                duplicates.append(isbn)

        self.log_duplicate_isbns(duplicates)

        return filtered_records

    def log_duplicate_isbns(self, duplicates: List[str]):
        """Log the list of duplicate ISBNs encountered.

        :param duplicates: List of duplicate ISBNs.
        """

        for isbn in duplicates:
            error_msg = f"ISBN {isbn} has duplicate product records in the ONIX data source."
            self.errors.append(error_msg)

    def get_pref_work_id(self, identifiers: List[Dict]) -> Tuple[Union[None, str], Union[None, str]]:
        """
        Tries to map the work identifier back to ISBN.
        :param identifiers: List of WorkIdentifiers.
        :return: Preferred work id type and the work id. None represents unknoown work_id type.
        """

        id_pref = {"ISBN-13": 0, "GTIN_13": 1, "DOI": 2, "Proprietary": 3}

        work_id = None
        work_id_type = None
        best_pref = float("inf")
        for wid in identifiers:
            wid_type = wid["WorkIDType"]
            pref = len(id_pref)

            if wid_type in id_pref:
                pref = id_pref[wid_type]

            if pref < best_pref:
                best_pref = pref
                work_id = wid["IDValue"]
                work_id_type = wid_type

        if work_id_type == "ISBN-13":
            return "ISBN13", work_id

        if work_id_type == "GTIN_13" and work_id in self.gtin13_to_product:
            product = self.gtin13_to_product[work_id]
            return "ISBN13", product["ISBN13"]

        if work_id_type == "Proprietary" and work_id in self.proprietary_to_product:
            product = self.proprietary_to_product[work_id]
            return "ISBN13", product["ISBN13"]

        return work_id_type, work_id

    def aggregate(self) -> List[BookWork]:
        """
        Run the aggregation process.
        Separate out the records into those containing RelatedWorks info, RelatedProducts info, and neithe.
        For a single publisher, this should only ever be 1 of the cases.
        Run different aggregation procedures for the 3 cases.
        If publishers are doing something funky, then we need to revisit this procedure.
        :return: List of BookWork objects representing the aggregated product records.
        """

        if self.n == 0:
            return self.works

        self.agg_relworks()
        self.agg_relproducts()
        self.works = self.get_works_from_partition(self.uf.get_partition())

        return self.works

    def is_relevant_work_relation(self, relwork: dict) -> bool:
        """
        Check if the work relation code indicates a manifestation.
        :param relwork: Related work.
        :return: Whether the related work is a manifestation of the current work.
        """

        code = relwork["WorkRelationCode"]
        if code == "Manifestation of":
            return True

        return False

    def log_agg_relworks_errors(self, pisbn: str, wtype: str, wid: str):
        """Log any errors from aggregating along RelatedWorks.
        :param pisbn: The product's ISBN.
        :param wtype: Type of WorkID.
        :param wid: WorkID of the related work.
        :return: True if we logged an error, False if it was OK.
        """

        if wid not in self.isbn13_to_index:
            error_msg = f"Product ISBN13:{pisbn} is a manifestation of {wtype}:{wid}, which is not given as a product identifier in any ONIX product record."
            self.errors.append(error_msg)
            return True

        return False

    def agg_relworks(self) -> List[BookWork]:
        """
        Collect the entries with "Manifestation of" relation codes into a single work.  The Work ID from that field will be used.  This assumes that every product
        record has a "Manifestation of" field entry that either points to itself or
        something else. Revisit this if a publisher does it differently.
        :return: List of BookWork objects that categorise the product records.
        """

        for record in self.records:
            if "RelatedWorks" not in record:
                continue

            isbn = record["ISBN13"]
            current_idx = self.isbn13_to_index[isbn]
            manifestations = set()

            for relwork in record["RelatedWorks"]:
                if not self.is_relevant_work_relation(relwork):
                    continue

                wtype, wid = self.get_pref_work_id(relwork["WorkIdentifiers"])
                if self.log_agg_relworks_errors(isbn, wtype, wid):
                    continue

                manifestations.add(wid)
                wid_idx = self.isbn13_to_index[wid]
                self.uf.unite(current_idx, wid_idx)

            if len(manifestations) > 1:
                self.log_get_works_lookup_table_errors(manifestations, isbn)

    def get_pid_idx(self, pid_type: str, pid: str) -> Union[None, int]:
        """
        Get the product index (in self.records) using the Product ID information.
        :param pid_type: Product ID type.
        :param pid: Product ID.
        :return: Index to the product or None if record doesn't exist.
        """

        if pid_type == "ISBN13":
            isbn = pid
        elif pid_type == "GTIN_13":
            if pid not in self.gtin13_to_product:
                return None
            relrec = self.gtin13_to_product[pid]
            isbn = relrec["ISBN13"]
        elif pid_type == "PID_Proprietary":
            if pid not in self.proprietary_to_product:
                return None
            relrec = self.proprietary_to_product[pid]
            isbn = relrec["ISBN13"]
        else:
            raise Exception(f"No handling implemented for {pid_type}")

        if isbn not in self.isbn13_to_index:
            return None

        return self.isbn13_to_index[isbn]

    def set_relevant_product_relation_codes(self) -> Set[str]:
        """
        :return: Set of relevant product codes indicating a manifestation of the current work.
        """

        return set(
            [
                "Alternative format",
                "Is remaindered as",
                "Is remainder of",
                "Epublication based on (print product)",
                "POD replacement for",
                "Replaced by POD",
                "Electronic version available as",
                "Is signed version of",
                "Has signed version",
            ]
        )

    def is_relevant_product_relation(self, relprod: dict) -> bool:
        """
        Check if the related product has a relation indicating it's the same work as the current work.
        :param relprod: Related product.
        :return: Whether the product is a manifestation.
        """

        codes = set(relprod["ProductRelationCodes"])
        for relcode in self.relevant_product_codes:
            if relcode in codes:
                return True

        return False

    def log_agg_related_product_errors(self, pisbn: str, relation: str, ptype: str, pid: str):
        """
        Log the errors from aggregating along RelatedProducts.

        :param pisbn: The product's ISBN.
        :param relation: The relation code.
        :param ptype: Related product's identifer type.
        :param pid: Related product's identifier.
        """

        error_msg = f"Product ISBN13:{pisbn} has a related product {ptype}:{pid} of types {relation} which is not given as a product identifier in any ONIX product record."
        self.errors.append(error_msg)

    def get_works_from_partition(self, partition: List[List[int]]) -> List[BookWork]:
        """
        Convert the partition of equivalence classes of record indices into equivalence classes of BookWork objects.
        :param partition: Partition of product record indices as works equivalence classes.
        :return: Partition as BookWork objects.
        """

        n = len(self.records)
        index_to_product = {i: self.records[i] for i in range(n)}

        works = list()
        for part in partition:
            products = [index_to_product[i] for i in part]
            # First ISBN is the WorkID representative
            work_id = products[0]["ISBN13"]
            works.append(BookWork(work_id=work_id, work_id_type="ISBN13", products=products))

        return works

    def agg_relproducts(self) -> List[BookWork]:
        """
        Aggregate the entries with targeted relation codes into a single work.
        Currently this is:
            "Alternative format".
        The Work ID will be an arbitrary ISBN13 representative from each work.
        :return: List of BookWork objects that categorise the product records.
        """

        for record in self.records:
            if "RelatedProducts" not in record:
                continue

            isbn = record["ISBN13"]
            current_idx = self.isbn13_to_index[isbn]

            for relprod in record["RelatedProducts"]:
                if not self.is_relevant_product_relation(relprod):
                    continue

                pid_type, pid = get_pref_product_id(relprod)
                pid_idx = self.get_pid_idx(pid_type, pid)

                if pid_idx is None:
                    relation = relprod["ProductRelationCodes"]
                    self.log_agg_related_product_errors(isbn, relation, pid_type, pid)
                    continue

                self.uf.unite(current_idx, pid_idx)

    def log_get_works_lookup_table_errors(self, manifestations: Set[str], isbn: str):
        """Log an error when an ISBN is assigned to multiple WorkIDs.

        :param manifestations: List of work IDs manifested by an ISBN.
        :param isbn: ISBN that has multiple WorkID assignments.
        """

        error_msg = f"Warning: product {isbn} has multiple work ID assignments: {manifestations}"
        self.errors.append(error_msg)

    def get_works_lookup_table(self) -> List[dict]:
        """
        Aggregate the products into works, and output a list of dicts ready for jsonline conversion and BQ loading.
        Keys: ISBN, Work ID.

        :return: List of dicts.
        """

        lookup_entries = list()

        for work in self.works:
            for isbn in work.isbns:
                mapping = {"isbn13": str(isbn), "work_id": str(work.work_id)}
                lookup_entries.append(mapping)

        return lookup_entries


class BookWorkFamilyAggregator:
    """
    Aggregates different editions of works into a family.  The methodology will be similar to BookWorkAggregator.
    This works with lists of BookWork objects rather than product records, so you need to have already aggregated
    products to works.
    """

    def __init__(self, works: List[BookWork]):
        """
        :param works: List of work objects.
        """

        self.works = works
        self.relevant_product_codes = self.set_relevant_product_codes()
        self.work_families = list()

    def set_relevant_product_codes(self) -> Set[str]:
        """
        :return: Set of relevant product codes indicating different editions.
        """

        return set(
            [
                "Replaces",
                "Replaced by",
                "Is special edition of",
                "Has special edition",
                "Is prebound edition of",
                "Is original of prebound edition",
                "Is facsimile of",
                "Is original of facsimile",
                "Is later edition of first edition",
            ]
        )

    def aggregate(self) -> List[BookWorkFamily]:
        """
        Run the aggregation process.
        Things that hint at edition information:
          1. "Replaces", "Replaced by", "Is later edition of first edition" relation in RelatedProducts.
          2. [Not implemented] "Derived from" and "Related work is derived from this" relation in RelatedWorks.
             This might need to be supplemented with other info.
          3. [Not implemented] Title, Authors, EditionNumber.
        :return: List of book work families.
        """

        self.work_families = self.agg_relproducts()
        return self.work_families

    def get_identifier_to_index_table(self, identifier: str) -> Dict:
        """
        Create a lookup table mapping identifiers to the index of the works list.
        :param identifier: Identifier name, e.g., ISBN13.
        :return: Lookup table.
        """

        n = len(self.works)
        id_to_index = {}

        for i in range(n):
            work = self.works[i]

            for product in work.products:
                if identifier in product:
                    pid = product[identifier]
                    id_to_index[pid] = i

        return id_to_index

    def get_wid_idx(
        self, pid_type: str, pid: str, isbn13_to_index: dict, gtin13_to_index: dict, proprietary_to_index: dict
    ) -> Union[None, int]:
        """
        Get the index into the works list for the product id.

        :param pid_type: Product identifier type.
        :param pid: Product ID.
        :param isbn13_to_index: ISBN lookup table.
        :param gtin13_to_index: GTIN13 lookup table.
        :param proprietary_to_index: Proprietary ID lookup table.
        :return: Index for the work in the works list, or None if the record doesn't exist.
        """

        if pid_type == "ISBN13":
            if pid not in isbn13_to_index:
                return None
            return isbn13_to_index[pid]
        elif pid_type == "GTIN_13":
            if pid not in gtin13_to_index:
                return None
            return gtin13_to_index[pid]
        elif pid_type == "PID_Proprietary":
            if pid not in proprietary_to_index:
                return None
            return proprietary_to_index[pid]
        else:
            raise Exception(f"No handling implemented for {pid_type}")

    def is_relevant_product_relation(self, relprod: dict) -> bool:
        """
        Check whether a related product contains a code indicating different (equivalent content) editions.
        :param relprod: Related product.
        :return: Whether the related product is a different edition.
        """

        codes = set(relprod["ProductRelationCodes"])

        for relcode in self.relevant_product_codes:
            if relcode in codes:
                return True

        return False

    def link_related_products(self) -> List[List[int]]:
        """
        Partition the works into equivalence classes of work families based on product relation codes.
        :return: Partition of the works into work families (using work indices of the works list).
        """

        n = len(self.works)
        gtin13_to_index = self.get_identifier_to_index_table("GTIN_13")
        isbn13_to_index = self.get_identifier_to_index_table("ISBN13")
        proprietary_to_index = self.get_identifier_to_index_table("PID_Proprietary")

        uf = UnionFind(n)
        for work_idx in range(n):
            work = self.works[work_idx]
            for product in work.products:
                for relprod in product["RelatedProducts"]:
                    if not self.is_relevant_product_relation(relprod):
                        continue

                    pid_type, pid = get_pref_product_id(relprod)
                    wid = self.get_wid_idx(pid_type, pid, isbn13_to_index, gtin13_to_index, proprietary_to_index)

                    # BookWorkAggregator should be logging all errors. No need for extra logging here.
                    if wid is None:
                        continue

                    uf.unite(work_idx, wid)

        return uf.get_partition()

    def agg_relproducts(self) -> List[BookWorkFamily]:
        """
        Collect the entries with "Replaces", "Replaced by", "Is later edition of first edition" relation codes into a single family.  The Work Family ID will be an
        arbitrary WorkID representative.
        :return: List of BookWork objects that categorise the product records.
        """

        partition = self.link_related_products()
        workfamilies = list()

        for part in partition:
            works = [self.works[i] for i in part]
            work_family_id = works[0].work_id  # Take a representative
            work_family_id_type = works[0].work_id_type
            workfamilies.append(
                BookWorkFamily(works=works, work_family_id=work_family_id, work_family_id_type=work_family_id_type)
            )

        return workfamilies

    def get_works_family_lookup_table(self) -> Dict:
        """
        Aggregate the works into work families, and output a list of dicts ready for jsonline conversion and BQ loading.
        Keys: ISBN, Work family ID.

        :return: List of dicts.
        """

        lookup_entries = list()

        for work_family in self.work_families:
            for work in work_family.works:
                for isbn in work.isbns:
                    mapping = {"isbn13": str(isbn), "work_family_id": str(work_family.work_family_id)}
                    lookup_entries.append(mapping)

        return lookup_entries



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# sys.path.insert(0, os.path.abspath('.'))
import sys
import os
import shutil
from pathlib import Path

from recommonmark.transform import AutoStructify

from generate_schema_csv import generate_csv, generate_csv_pdf, generate_latest_files

# -- Project information -----------------------------------------------------

project = "Book Usage Data Workflows"
copyright = "2020-2022 Curtin University"
author = "Curtin University"


# -- Options for PDFL output -------------------------------------------------

latex_elements = {
    'sphinxsetup': "verbatimforcewraps, verbatimmaxunderfull=0",
    'extraclassoptions': 'openany,oneside',
    'preamble': r'''
        \usepackage[none]{hyphenat}
        \usepackage{makeidx}
        \makeindex
        \makeatletter
        \renewenvironment{theindex}
        {\setlength{\parindent}{0pt}\raggedright\small\let\item\@idxitem}
        {\makeatother}
        \usepackage{etoolbox}
        \patchcmd{\sphinxverbatimintable}{\sphinxsetup{VerbatimColor=\sphinxbaseblack\sphinxtightlistings}}{\sphinxsetup{VerbatimColor=\sphinxbaseblack\sphinxtightlistings\raggedright\scriptsize}}{}{}
    '''
}


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_rtd_theme",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "autoapi.extension",
    "recommonmark",
    "sphinx.ext.autodoc",
]

autodoc_typehints = 'description'

# Auto API settings: https://github.com/readthedocs/sphinx-autoapi
autoapi_type = "python"
autoapi_dirs = ["../oaebu_workflows"]
autoapi_add_toctree_entry = False
autoapi_python_use_implicit_namespaces = True
autoapi_root = "oaebu_workflows/api"

autoapi_python_use_implicit_namespaces = True
autoapi_python_class_content = 'both'

# Add any paths that contain templates here, relative to this directory.
templates_path = ["templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_logo = "logo.png"
html_theme_options = {
    "logo_only": True,
    "display_version": False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []


# recommonmark config, used to enable rst to be evaluated within markdown files
def setup(app):
    app.add_config_value("recommonmark_config", {"enable_eval_rst": True, "auto_toc_tree_section": "Contents"}, True)
    app.add_transform(AutoStructify)

# -- Build options to format tables for html or pdf output -------------------------------------------------

# Determine the command used to build the documentation
build_command = ' '.join(sys.argv)

if 'html' in build_command:
    generate_csv(schema_dir="../oaebu_workflows/database/schema")
    generate_latest_files()
    html_build_dir = "_build/html"
    Path(html_build_dir).mkdir(exist_ok=True, parents=True)
elif 'latexpdf' in build_command:
   generate_csv_pdf(schema_dir="../oaebu_workflows/database/schema")
   generate_latest_files()
   latex_build_dir = "_build/latex"
   Path(latex_build_dir).mkdir(exist_ok=True, parents=True)
else:
    generate_csv(schema_dir="../oaebu_workflows/database/schema")
    generate_latest_files()
    html_build_dir = "_build/html"
    Path(html_build_dir).mkdir(exist_ok=True, parents=True)



# Copyright 2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Keegan Smith

import os
import re
import json
import shutil
import subprocess
import logging
import xmltodict
import tempfile
from tempfile import TemporaryDirectory
from glob import glob
from typing import List, Tuple, Literal, Union, Mapping, Any
from dataclasses import dataclass
from copy import deepcopy

from onixcheck import validate as validate_onix

from oaebu_workflows.config import schema_folder
from observatory.platform.config import observatory_home
from observatory.platform.utils.http_download import download_file
from observatory.platform.utils.proc_utils import wait_for_process
from observatory.platform.files import save_jsonl_gz, load_jsonl


@dataclass
class OnixParser:
    """Class for storing infromation on the java ONIX parser

    :param filename: The name of the java ONIX parser file
    :param url: The url to use for downloading the parser
    :param bash template: The path to the bash template "onix_parser.sh.jinja2"
    """

    filename = "coki-onix-parser-1.2-SNAPSHOT-shaded.jar"
    url = "https://github.com/The-Academic-Observatory/onix-parser/releases/download/v1.3.0/coki-onix-parser-1.2-SNAPSHOT-shaded.jar"
    cmd = "java -jar {parser_path} {input_dir} {output_dir}"


class OnixTransformer:
    def __init__(
        self,
        *,
        input_path: str,
        output_dir: str,
        filter_products: bool = False,
        error_removal: bool = False,
        normalise_related_products: bool = False,
        deduplicate_related_products: bool = False,
        elevate_related_products: bool = False,
        add_name_fields: bool = False,
        collapse_subjects: bool = False,
        filter_schema: str = os.path.join(schema_folder(workflow_module="oapen_metadata_telescope"), "oapen_metadata_filter.json"),
        invalid_products_name: str = "invalid_products.xml",
        save_format: Literal["json", "jsonl", "jsonl.gz"] = "jsonl.gz",
        keep_intermediate: bool = False,
    ) -> None:
        """
        Constructor for the MetadataTransformer class.

        :param input_path: The path to the metadata file
        :param output_dir: The directory to output the transformed metadata
        :param filter_products: Filter the metadata through a filter schema
        :param error_removal: Remove products containing errors
        :param normalise_related_products: Fix imporperly formatted related products
        :param deduplicate_related_products: Deduplicate related products
        :param add_name_fields: Add the Contributor.PersonName and Contributor.InvertedPersonName fields where possible
        :param collapse_subjects: Collapse subjects into semicolon-separated strings
        :param filter_schema: The filter schema to use. Required if filter_products is True
        :param invalid_products_name: The name of the invalid products file.
        :param save_format: The format to save the transformed metadata in - json, jsonl, or jsonl.gz
        :param keep_intermediate: Keep the intermediate files
        """
        self.input_path = input_path
        self.output_dir = output_dir
        self.filter_products = filter_products
        self.error_removal = error_removal
        self.normalise_related_products = normalise_related_products
        self.deduplicate_related_products = deduplicate_related_products
        self.elevate_related_products = elevate_related_products
        self.add_name_fields = add_name_fields
        self.collapse_subjects = collapse_subjects
        self.filter_schema = filter_schema
        self.invalid_products_name = invalid_products_name
        self.save_format = save_format
        self.keep_intermediate = keep_intermediate

        if self.filter_products and not self.filter_schema:
            raise ValueError("Product filtering requires a provided filter_schema")

        # Use this attribute to keep track of the last modified metadata file
        self._current_md_path = input_path

        # Temp directory for working in
        self._work_dir = tempfile.mkdtemp()

    @property
    def current_metadata(self) -> Union[List[dict], Mapping[str, Any]]:
        return self._load_metadata(self._current_md_path)

    def __del__(self):
        # Remove the temporary directory
        if hasattr(self, "_work_dir") and self._work_dir:
            shutil.rmtree(self._work_dir, ignore_errors=True)

    def transform(self):
        """
        Transform the oapen metadata XML file based on the supplied options.

        The transformations will be done in the following order. Transforms not included will be skipped:
        1) Filter the XML metadata using a schema to keep the desired fields only
        2) Remove remaining products containing errors
        3) Fix incorrectly formatted related products
        4) Elevate related products to the product level
        5) Construct the Contributor.PersonName and Contributor.InvertedPersonName fields where possible
        6) Parse through the java parser to return .jsonl format - This is always done
        7) Collapse subjects into semicolon-separated strings
        """
        settings = []
        if self.filter_products:
            settings.append("Filter Products")
        if self.error_removal:
            settings.append("Error Removal")
        if self.normalise_related_products:
            settings.append("Normalise Related Products")
        if self.deduplicate_related_products:
            settings.append("Deduplicate Related Products")
        if self.elevate_related_products:
            settings.append("Elevate Related Products")
        settings.append("Parse ONIX")
        if self.add_name_fields:
            settings.append("Add Name Fields")
        if self.collapse_subjects:
            settings.append("Collapse Subjects")
        logging.info("Applying transformation in the following order:")
        logging.info(" | ".join(settings))

        if self.filter_products:
            self._filter_products()
        if self.error_removal:
            self._remove_errors()
        if self.normalise_related_products:
            self._normalise_related_products()
        if self.deduplicate_related_products:
            self._deduplicate_related_products()
        if self.elevate_related_products:
            self._elevate_related_products()
        self._apply_parser()
        if self.add_name_fields:
            self._apply_name_fields()
        if self.collapse_subjects:
            self._collapse_subjects()
        self._save_metadata(self.current_metadata, os.path.join(self.output_dir, f"transformed.{self.save_format}"))
        logging.info(f"Saved transformed metadata to {self._current_md_path}")
        return self._current_md_path

    def _intermediate_file_path(self, file_name):
        dir_ = self.output_dir if self.keep_intermediate else self._work_dir
        return os.path.join(dir_, file_name)

    def _save_metadata(self, metadata: Union[List[dict], Mapping[str, Any]], file_path: str):
        save_path = os.path.join(self._work_dir, file_path)
        format = re.search(r"\.(.*)$", file_path).group(1)
        if format == "xml":
            if not isinstance(metadata, Mapping):
                raise TypeError(f"Metadata must be of type Mapping, instead got type {type(metadata)}")
            with open(save_path, "w") as f:
                xmltodict.unparse(metadata, output=f, pretty=True)
        elif format == "json":
            with open(save_path, "w") as f:
                json.dump(metadata, f)
        elif format == "jsonl":
            with open(save_path, "w") as f:
                for m in metadata:
                    json.dump(m, f)
                    f.write("\n")
        elif format == "jsonl.gz":
            if not type(metadata) == list:
                raise TypeError(f"Metadata must be of type list, instead got type {type(metadata)}")
            save_jsonl_gz(save_path, metadata)
        else:
            raise ValueError(f"Unsupported format: {format}")
        self._current_md_path = save_path

    def _load_metadata(self, file_path: str):
        format = re.search(r"\.(.*)$", file_path).group(1)
        if format == "xml":
            with open(file_path, "rb") as f:
                metadata = xmltodict.parse(f)
        elif format == "json":
            with open(file_path, "r") as f:
                metadata = json.load(f)
        elif format == "jsonl" or format == "jsonl.gz":
            metadata = load_jsonl(file_path)
        else:
            raise ValueError(f"Unsupported format: {format}")
        return metadata

    def _filter_products(self):
        with open(self.filter_schema, "r") as f:
            schema = json.load(f)
        filtered_metadata = filter_through_schema(self.current_metadata, schema)
        self._save_metadata(filtered_metadata, self._intermediate_file_path("filtered.xml"))

    def _remove_errors(self):
        invalid_products_path = os.path.join(self.output_dir, self.invalid_products_name)
        save_path = self._intermediate_file_path("errors_removed.xml")
        remove_invalid_products(self._current_md_path, save_path, invalid_products_path)
        self._current_md_path = save_path

    def _normalise_related_products(self):
        metadata = deepcopy(self.current_metadata)
        if not isinstance(metadata, Mapping):
            raise TypeError(f"Metadata must be of type Mapping, instead is type {type(metadata)}")
        fixed_products = normalise_related_products(metadata["ONIXMessage"]["Product"])
        metadata["ONIXMessage"]["Product"] = fixed_products

        self._save_metadata(metadata, self._intermediate_file_path("normalised.xml"))

    def _deduplicate_related_products(self):
        metadata = deepcopy(self.current_metadata)
        if not isinstance(metadata, Mapping):
            raise TypeError(f"Metadata must be of type Mapping, instead is type {type(metadata)}")
        deduplicated_products = deduplicate_related_products(metadata["ONIXMessage"]["Product"])
        metadata["ONIXMessage"]["Product"] = deduplicated_products
        self._save_metadata(metadata, self._intermediate_file_path("deduplicated.xml"))

    def _elevate_related_products(self):
        metadata = deepcopy(self.current_metadata)
        if not isinstance(metadata, Mapping):
            raise TypeError(f"Metadata must be of type Mapping, instead is type {type(metadata)}")
        logging.info(f"Original product count: {len(metadata['ONIXMessage']['Product'])}")
        elevated = elevate_related_products(metadata["ONIXMessage"]["Product"])
        logging.info(f"Product count after elevating related products: {len(elevated)}")
        metadata["ONIXMessage"]["Product"] = elevated
        self._save_metadata(metadata, self._intermediate_file_path("elevated.xml"))

    def _apply_parser(self):
        success, parser_path = onix_parser_download()
        if not success:
            raise RuntimeError("Failed to download parser")

        with TemporaryDirectory() as tmp_dir:
            shutil.copy(self._current_md_path, os.path.join(tmp_dir, "input.xml"))
            success = onix_parser_execute(parser_path, input_dir=tmp_dir, output_dir=self.output_dir)
        if not success:
            raise RuntimeError("Failed to execute parser")

        self._current_md_path = os.path.join(self.output_dir, "full.jsonl")

    def _apply_name_fields(self):
        metadata = deepcopy(self.current_metadata)
        if not isinstance(metadata, list):
            raise TypeError(f"Metadata must be of type list, instead is type {type(metadata)}")
        metadata = create_personname_fields(metadata)
        self._save_metadata(metadata, self._intermediate_file_path("name_applied.jsonl"))

    def _collapse_subjects(self):
        metadata = deepcopy(self.current_metadata)
        if not isinstance(metadata, list):
            raise TypeError(f"Metadata must be of type list, instead is type {type(metadata)}")
        metadata = collapse_subjects(metadata)
        self._save_metadata(metadata, self._intermediate_file_path("collapsed.jsonl"))


def onix_parser_download(download_dir: str = observatory_home("bin")) -> Tuple[bool, str]:
    """Downloads the ONIX parser from Github

    :param download_dir: The directory to download the file to
    :return: (Whether the download operation was a success, The (expected) location of the downloaded file)
    """
    success, _ = download_file(url=OnixParser.url, prefix_dir=download_dir, filename=OnixParser.filename)
    parser_path = os.path.join(download_dir, OnixParser.filename)
    success = os.path.isfile(parser_path) if success else success  # Second check for file existence
    return success, parser_path


def onix_parser_execute(parser_path: str, input_dir: str, output_dir: str) -> bool:
    """Executes the Java ONIX parser. Requires a .xml file in the input directory.

    :param parser_path: Filepath of the parser
    :param input_dir: The input directory - first argument of the parser
    :param output_dir: The output directory - second argument of the parser
    :return: Whether the task succeeded or not (return code 0 means success)
    """
    if not glob(os.path.join(input_dir, "*.xml")):
        logging.error(f"No .xml file found in input directory: {input_dir}")
        return False

    cmd = OnixParser.cmd.format(parser_path=parser_path, input_dir=input_dir, output_dir=output_dir)
    process = subprocess.Popen(cmd.split(" "), shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = wait_for_process(process)
    if stdout:
        logging.info(stdout)
    if stderr:
        logging.info(stderr)
    if process.returncode != 0:
        logging.error(f"Bash command failed `{cmd}`: {stderr}")
        return False
    if not os.path.exists(os.path.join(output_dir, "full.jsonl")):  # May fail to produce file even with returncode=0
        logging.error(f"No .jsonl file found in output directory: {output_dir}")
        return False

    return True


def collapse_subjects(onix: List[dict]) -> List[dict]:  # TODO: alter this to work with pre-parsed onix (.xml)
    """The book product table creation requires the keywords (under Subjects.SubjectHeadingText) to occur only once
    Some ONIX feeds return all keywords as separate entires. This function finds and collapses each keyword into a
    semi-colon separated string. Other common separators will be replaced with semi-colons.

    :param onix: The onix feed
    :return: The onix feed after collapsing the keywords of each row
    """
    for row in onix:
        # Create the joined keywords in this row
        keywords = []
        for subject in row["Subjects"]:
            if subject["SubjectSchemeIdentifier"] != "Keywords":
                continue
            subject_heading_text = [i for i in subject["SubjectHeadingText"] if i is not None]  # Remove Nones
            if not subject_heading_text:  # Empty list
                continue
            keywords.append("; ".join(subject_heading_text))
        # Enforce a split by semicolon
        keywords = "; ".join(keywords)
        keywords = keywords.replace(",", ";")
        keywords = keywords.replace(":", ";")

        # Replace one of the subrows with the new keywords string
        keywords_replaced = False
        remove_indexes = []
        for i, subject in enumerate(row["Subjects"]):
            if subject["SubjectSchemeIdentifier"] == "Keywords":
                if not keywords_replaced:
                    subject["SubjectHeadingText"] = [keywords]
                    keywords_replaced = True
                else:
                    remove_indexes.append(i)

        # Remove additional "keywords" subrows
        for i in sorted(remove_indexes, reverse=True):
            del row["Subjects"][i]

    return onix


def create_personname_fields(onix_products: List[dict]) -> List[dict]:
    """Given an ONIX product list, attempts to populate the Contributors.PersonName and/or Contributors.PersonNameInverted
    fields by concatenating the Contributors.NamesBeforeKey and Contributors.KeyNames fields where possible

    :param onix: The input onix feed
    :return: The onix feed with the additional fields populated where possible
    """

    def _can_make_personname(contributor: dict) -> Union[bool, None]:
        return contributor.get("KeyNames") and contributor.get("NamesBeforeKey")

    def _can_make_personname_inverted(contributor: dict) -> Union[bool, None]:
        return contributor.get("NamesBeforeKey") and contributor.get("KeyNames")

    return_products = deepcopy(onix_products)
    for product in return_products:
        try:
            contributors = product["Contributors"]
        except KeyError:
            continue

        if not isinstance(contributors, list):
            contributors = [contributors]

        for c in contributors:
            if _can_make_personname(c) and not c.get("PersonName"):
                c["PersonName"] = f"{c['NamesBeforeKey']} {c['KeyNames']}"
            if _can_make_personname_inverted(c) and not c.get("PersonNameInverted"):
                c["PersonNameInverted"] = f"{c['KeyNames']}, {c['NamesBeforeKey']}"
    return return_products


def elevate_product_identifiers(related_product: dict) -> List[dict]:
    """Given a single <RelatedProduct>, returns a list of <RelatedProduct> elements by elevating
    <ProductIdentifier> elements that shouldn't be there.
    A <ProductIdentifier> element should only appear once in the <RelatedProduct> unless it has the same
    <IDValue> and a different <ProductIDType>

    <RelatedProduct>
        <ProductRelationCode></ProductRelationCode>
        <ProductIdentifier>
            <IDValue></IDValue>
            <ProductIDType></ProductIDType>
        </ProductIdentifier>
    </RelatedProduct>

    :param related_product: The single <RelatedProduct> element with product identifiers to elevate
    :return: A list of <RelatedProduct> elements
    """
    product_identifiers = _get_product_identifiers(related_product)

    # Find all unique <ProductIdentifier> elements. Elements with same ID and different ID types will be put in the
    # same <RelatedProduct> element.
    product_contains = {}  # key : id_value, value : set(id_types)
    for product_identifier in product_identifiers:
        id_type = product_identifier["ProductIDType"]
        id_value = product_identifier["IDValue"]
        if id_value not in product_contains.keys():
            product_contains[id_value] = set()
            product_contains[id_value].add(id_type)
        else:
            product_contains[id_value].add(id_type)

    new_related_products = []
    product_relation_code = related_product["ProductRelationCode"]
    for product_identifier, id_types in product_contains.items():
        identifier_list = [{"ProductIDType": id_type, "IDValue": product_identifier} for id_type in id_types]
        product = {"ProductRelationCode": product_relation_code, "ProductIdentifier": identifier_list}
        new_related_products.append(product)
    return new_related_products


def normalise_related_products(onix_products: List[dict]) -> List[dict]:
    """Many products have incorrect related products. The related products should be a list of <RelatedProduct> but is
    instead a single <RelatedProduct> with many <ProductIdentifier> elements.
    This function fixes this issue by creating a new <RelatedProduct> for each <ProductIdentifier> where necessary.
    Note that a <RelatedProduct> can still have more than one <ProductIdentifier> if the <IDValue> elements are the
    same, but the <ProductIDType> is different.

    :param product: The list of onix products to fix
    :return: The amended products
    """
    return_products = deepcopy(onix_products)
    for product in return_products:
        related_products = _get_related_products(product)
        if not related_products:
            continue

        new_related_products = []
        for related_product in related_products:
            new_related_products.extend(elevate_product_identifiers(related_product))
        product["RelatedMaterial"]["RelatedProduct"] = deepcopy(new_related_products)
    return return_products


def deduplicate_related_products(onix_products: List[dict]) -> List[dict]:
    """Removes any duplicated <RelatedProduct> elements. Will also remove parent ISBNs present in the <RelatedProduct>
    as they are effectively duplicates

    :param product: The list of onix products to fix
    :return: The amended products
    """
    return_products = deepcopy(onix_products)
    for product in return_products:
        seen_ids = {}
        id_relation_codes = {}
        product_isbn = _get_product_isbn(product)
        related_products = _get_related_products(product)
        if not related_products:
            continue

        # Find all unique <RelatedProduct> elements
        for related_product in related_products:
            for product_identifier in _get_product_identifiers(related_product):
                rp_id = product_identifier["IDValue"]
                rp_id_type = product_identifier["ProductIDType"]
                if rp_id == product_isbn and rp_id_type == "15":  # Ignore parent ISBN in related products
                    continue
                if rp_id not in seen_ids.keys():
                    seen_ids[rp_id] = set()
                seen_ids[rp_id].add(rp_id_type)
                id_relation_codes[rp_id] = related_product["ProductRelationCode"]

        # Rebuild the related products
        new_related_products = []
        for id, id_types in seen_ids.items():
            related_product = {"ProductRelationCode": id_relation_codes[id], "ProductIdentifier": []}
            for id_type in id_types:
                related_product["ProductIdentifier"].append({"ProductIDType": id_type, "IDValue": id})
            new_related_products.append(related_product)
        product["RelatedMaterial"]["RelatedProduct"] = deepcopy(new_related_products)

    return return_products


def elevate_related_products(onix_products: List[dict]) -> List[dict]:
    """Makes a copy of an ONIX product for each of its unique related products with an ISBN.
    The copies will swap the original ISBN with their own ISBN and make a unique record reference.
    This "elevates" all related products to the product level.
    Note that if the product identifier for the related product is not an ISBN, it will not be elevated

    Related Product Structure:
    <RelatedProduct>
        <ProductRelationCode></ProductRelationCode>
        <ProductIdentifier>
            <IDValue></IDValue>
            <ProductIDType></ProductIDType>
        </ProductIdentifier>
        <ProductIdentifier>
            <IDValue></IDValue>
            <ProductIDType></ProductIDType>
        </ProductIdentifier>
    </RelatedProduct>

    :param product: The ONIX product list
    :return: A list containing the original ONIX product and its related children products.
    """

    return_products = deepcopy(onix_products)

    # Get a list of product ISBNS
    parent_isbns = [_get_product_isbn(product) for product in return_products]

    # Get the original product ISBN
    added_products = []
    for product in return_products:
        related_products = _get_related_products(product)
        product_isbn = _get_product_isbn(product)

        if not product_isbn:
            continue  # Only elevate if parent product has an ISBN

        for related_product in related_products:
            relation_code = related_product["ProductRelationCode"]
            if relation_code != "06":
                continue  # We expect only "Alternative Format" relations - https://onix-codelists.io/codelist/51

            # Find the ISBN for the related product
            rp_isbn = _get_product_isbn(related_product)
            if not rp_isbn:
                continue  # No ISBNs in related product
            if rp_isbn in parent_isbns:
                continue  # No need to elevate as this is a duplicate of one of the original ISBNs

            # Copy & make modifications to product
            new_product = deepcopy(product)
            new_product["ProductIdentifier"] = {"ProductIDType": "15", "IDValue": rp_isbn}
            new_product["RecordReference"] += f"_{rp_isbn}"
            # Find and replace the original related product with the parent's ISBN
            for rp in _get_related_products(new_product):
                relation_code = rp["ProductRelationCode"]
                if relation_code != "06":
                    continue
                if _get_product_isbn(rp) == rp_isbn:
                    rp["ProductIdentifier"] = {"ProductIDType": "15", "IDValue": product_isbn}

            added_products.append(new_product)
            parent_isbns.append(rp_isbn)
    return return_products + added_products


def _get_product_isbn(product: dict) -> Union[str, None]:
    """Finds the ISBN of a product or relatedproduct"""
    product_identifiers = product["ProductIdentifier"]
    if not isinstance(product_identifiers, list):
        product_identifiers = [product_identifiers]
    for product_identifier in product_identifiers:
        if str(product_identifier["ProductIDType"]) == "15":
            return product_identifier["IDValue"]


def _get_related_products(product: dict):
    """Finds the related products of a product"""
    try:
        related_products = product["RelatedMaterial"]["RelatedProduct"]
    except (KeyError, TypeError):  # Product has no related products
        return []
    # If there's only one related product, it's not a list
    if not isinstance(related_products, list):
        related_products = [related_products]
    return related_products


def _get_product_identifiers(product: dict):
    """Finds the product identifiers of a product"""
    try:
        product_identifiers = product["ProductIdentifier"]
    except KeyError:  # Product has no product identifiers
        return []
    # If there's only one product identifier, it's not a list
    if not isinstance(product_identifiers, list):
        product_identifiers = [product_identifiers]
    return product_identifiers


@dataclass
class OnixProduct:
    """Represents a single ONIX product and its identifying reference for simplicity"""

    product: dict
    record_reference: str


def find_onix_product(all_lines: list, line_index: int) -> OnixProduct:
    """Finds the range of lines encompassing a <Product> tag, given a line_number that is contained in the product

    :param all_lines: All lines in the onix file
    :param line_number: The line number associated with the product
    :return: A two-tuple of the start and end line numbers of the product
    :raises ValueError: Raised if the return would encompass a negative index, indicating the input line was not in a product
    """
    if line_index < 0 or line_index >= len(all_lines):
        raise IndexError(f"Supplied line index {line_index} is not within the length of the file: {len(all_lines)}")

    # Go up until we find <Product>
    begin_index = line_index
    begin_content = all_lines[begin_index]
    while not "<Product>" in begin_content:
        begin_index -= 1
        if begin_index < 0:
            raise ValueError(f"Product not found surrounding line {line_index}")
        begin_content = all_lines[begin_index]

    # Go up until we find </Product>
    finish_index = line_index
    finish_content = all_lines[finish_index]
    while not "</Product>" in finish_content:
        finish_index += 1
        if finish_index >= len(all_lines):
            raise ValueError(f"Product not found surrounding line {line_index}")
        finish_content = all_lines[finish_index]

    product_str = "".join(all_lines[begin_index : finish_index + 1])
    product = xmltodict.parse(product_str)["Product"]
    if not product:
        raise ValueError(f"Product field is empty for product at line {begin_index}")
    record_reference = product.get("RecordReference")
    if not record_reference:
        raise KeyError(f"No RecordReference for product: {product_str}")

    return OnixProduct(product, record_reference)


def remove_invalid_products(input_xml: str, output_xml: str, invalid_products_file: str = None) -> None:
    """Attempts to validate the input xml as an ONIX file. Will remove any products that contain errors.

    :param input_xml: The filepath of the xml file to validate
    :param output_xml: The output filepath
    :param invalid_products_file: The filepath to write the invalid products to. Ignored if unsupplied.
    """
    # Get the line numbers of any errors
    errors = validate_onix(input_xml)
    error_lines = [int(e.location.split(":")[-2]) - 1 for e in errors]

    # Ingest the file into a list and find onix products with errors
    with open(input_xml, "r") as f:
        metadata = f.readlines()

    invalid_products = [find_onix_product(metadata, line_number) for line_number in error_lines]
    invalid_references = set([product.record_reference for product in invalid_products])
    logging.info(
        f"Metadata feed has been trimmed and {len(invalid_references)} errors remain. Products with errors will be removed"
    )

    # Parse the xml to dictionary
    with open(input_xml, "rb") as f:
        metadata = xmltodict.parse(f)

    # Remove products matching the record references
    metadata["ONIXMessage"]["Product"] = [
        p for p in metadata["ONIXMessage"]["Product"] if p["RecordReference"] not in invalid_references
    ]

    # Create the clean XML file
    with open(output_xml, "w") as f:
        xmltodict.unparse(metadata, output=f, pretty=True)

    # Create the invalid product file
    if invalid_products_file:
        with open(invalid_products_file, "w") as f:
            metadata["ONIXMessage"]["Product"] = [p.product for p in invalid_products]
            xmltodict.unparse(metadata, output=f, pretty=True)
        logging.info(f"Invalid products written to {invalid_products_file}")


def filter_through_schema(input: Union[dict, list], schema: dict):
    """
    This function recursively traverses the input dictionary and compares it to the provided schema.
    It retains only the fields and values that exist in the schema structure, and discards
    any fields that do not match the schema.

    # Example usage with a dictionary and schema:
        input_dict = {
            "name": "John",
            "age": 30,
            "address": {
                "street": "123 Main St",
                "city": "New York",
                "zip": "10001"
            }
        }
        schema = {
            "name": null,
            "age": null,
            "address": {
                "street": null,
                "city": null
            }
        }
        filtered_dict = filter_dict_by_schema(input_dict, schema)
        filtered_dict will be:
        {
            "name": "John",
            "age": 30,
            "address": {
                "street": "123 Main St",
                "city": "New York"
            }
        }

    :param input: The dictionary to filter
    :param schema: The schema describing the desired structure of the dictionary
    """
    if isinstance(input, dict) and isinstance(schema, dict):
        return {key: filter_through_schema(value, schema.get(key)) for key, value in input.items() if key in schema}
    elif isinstance(input, list) and isinstance(schema, list):
        return [filter_through_schema(item, schema[0]) for item in input]
    else:
        return input



# Copyright 2020-2023 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs

import json
import os
from unittest.mock import ANY, MagicMock, patch
from urllib.parse import quote
from requests import Response

import httplib2
import httpretty
import pendulum
from airflow.models import Connection
from airflow.exceptions import AirflowException, AirflowSkipException
from airflow.utils.state import State
from click.testing import CliRunner
from googleapiclient.discovery import build
from googleapiclient.http import RequestMockBuilder

from oaebu_workflows.config import test_fixtures_folder
from oaebu_workflows.oaebu_partners import partner_from_str
from oaebu_workflows.irus_oapen_telescope.irus_oapen_telescope import (
    IrusOapenRelease,
    IrusOapenTelescope,
    call_cloud_function,
    cloud_function_exists,
    create_cloud_function,
    upload_source_code_to_bucket,
)
from observatory.platform.api import get_dataset_releases
from observatory.platform.observatory_config import CloudWorkspace, Workflow
from observatory.platform.gcs import gcs_blob_name_from_path, gcs_upload_file
from observatory.platform.bigquery import bq_table_id
from observatory.platform.observatory_environment import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    find_free_port,
    random_id,
)


class TestIrusOapenTelescope(ObservatoryTestCase):
    """Tests for the Oapen Irus Uk telescope"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.
        :param args: arguments.
        :param kwargs: keyword arguments.
        """
        super(TestIrusOapenTelescope, self).__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")
        self.publisher_name_v4 = (quote("UCL Press"),)
        self.publisher_uuid_v5 = ("df73bf94-b818-494c-a8dd-6775b0573bc2",)
        self.download_path = os.path.join(
            test_fixtures_folder(workflow_module="irus_oapen_telescope"), "download.jsonl.gz"
        )

    def test_dag_structure(self):
        """Test that the Oapen Irus Uk DAG has the correct structure."""
        dag = IrusOapenTelescope(
            dag_id="irus_oapen_test_dag",
            cloud_workspace=self.fake_cloud_workspace,
            publisher_name_v4=self.publisher_name_v4,
            publisher_uuid_v5=self.publisher_uuid_v5,
        ).make_dag()
        self.assert_dag_structure(
            {
                "check_dependencies": ["create_cloud_function"],
                "create_cloud_function": ["call_cloud_function"],
                "call_cloud_function": ["transfer"],
                "transfer": ["download_transform"],
                "download_transform": ["upload_transformed"],
                "upload_transformed": ["bq_load"],
                "bq_load": ["add_new_dataset_releases"],
                "add_new_dataset_releases": ["cleanup"],
                "cleanup": [],
            },
            dag,
        )

    def test_dag_load(self):
        """Test that the Oapen Irus Uk DAG can be loaded from a DAG bag."""

        env = ObservatoryEnvironment(
            workflows=[
                Workflow(
                    dag_id="irus_oapen_test",
                    name="My Oapen Irus UK Workflow",
                    class_name="oaebu_workflows.irus_oapen_telescope.irus_oapen_telescope.IrusOapenTelescope",
                    cloud_workspace=self.fake_cloud_workspace,
                    kwargs=dict(publisher_name_v4=self.publisher_name_v4, publisher_uuid_v5=self.publisher_uuid_v5),
                )
            ],
        )
        with env.create():
            self.assert_dag_load_from_config("irus_oapen_test")

    @patch("oaebu_workflows.irus_oapen_telescope.irus_oapen_telescope.build")
    @patch("oaebu_workflows.irus_oapen_telescope.irus_oapen_telescope.ServiceAccountCredentials")
    @patch("oaebu_workflows.irus_oapen_telescope.irus_oapen_telescope.AuthorizedSession.post")
    def test_telescope(self, mock_authorized_session, mock_account_credentials, mock_build):
        """Test the Oapen Irus Uk telescope end to end."""
        # Setup Observatory environment
        env = ObservatoryEnvironment(
            self.project_id, self.data_location, api_host="localhost", api_port=find_free_port()
        )

        # Setup Telescope
        execution_date = pendulum.datetime(year=2021, month=2, day=14)
        patner = partner_from_str("irus_oapen")
        patner.bq_dataset_id = env.add_dataset()
        telescope = IrusOapenTelescope(
            dag_id="irus_oapen_test",
            cloud_workspace=env.cloud_workspace,
            publisher_name_v4=self.publisher_name_v4,
            publisher_uuid_v5=self.publisher_uuid_v5,
            data_partner=patner,
        )
        # Fake oapen project and bucket
        IrusOapenTelescope.OAPEN_PROJECT_ID = env.project_id
        IrusOapenTelescope.OAPEN_BUCKET = random_id()

        # Mock the Google Cloud Functions API service
        mock_account_credentials.from_json_keyfile_dict.return_value = ""
        request_builder = RequestMockBuilder(
            {
                "cloudfunctions.projects.locations.functions.get": (
                    None,
                    json.dumps(
                        {
                            "name": "projects/project-id/locations/us-central1/functions/function-2",
                            "serviceConfig": {"uri": "https://oapen-access-stats-kkinbzaigla-ew.a.run.app"},
                        }
                    ),
                ),
                "cloudfunctions.projects.locations.functions.patch": (
                    None,
                    json.dumps(
                        {
                            "name": "projects/project-id/locations/us-central1/operations/d29ya2Z",
                            "done": True,
                            "response": {"message": "response"},
                        }
                    ),
                ),
            },
            check_unexpected=True,
        )
        mock_build.return_value = build(
            "cloudfunctions",
            "v2beta",
            cache_discovery=False,
            static_discovery=False,
            requestBuilder=request_builder,
        )

        dag = telescope.make_dag()

        # Create the Observatory environment and run tests
        with env.create(task_logging=True):
            with env.create_dag_run(dag, execution_date):
                # Use release to check results from tasks
                # release = IrusOapenRelease(
                #     dag_id=telescope.dag_id, run_id=env.dag_run.run_id, partition_date=execution_date.end_of("month")
                # )
                release = telescope.make_release(
                    run_id=env.dag_run.run_id,
                    data_interval_start=pendulum.parse(str(env.dag_run.data_interval_start)),
                    data_interval_end=pendulum.parse(str(env.dag_run.data_interval_end)),
                )[0]

                # Add airflow connections
                conn = Connection(conn_id=telescope.geoip_license_conn_id, uri="http://email_address:password@")
                env.add_connection(conn)
                conn = Connection(conn_id=telescope.irus_oapen_api_conn_id, uri="mysql://requestor_id:api_key@")
                env.add_connection(conn)
                conn = Connection(conn_id=telescope.irus_oapen_login_conn_id, uri="mysql://user_id:license_key@")
                env.add_connection(conn)

                # Test that all dependencies are specified: no error should be thrown
                ti = env.run_task(telescope.check_dependencies.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Test create cloud function task: no error should be thrown
                ti = env.run_task(telescope.create_cloud_function.__name__)
                self.assertEqual(ti.state, State.SUCCESS)

                # Test call cloud function task: no error should be thrown
                with httpretty.enabled():
                    # mock response of getting publisher uuid
                    url = "https://library.oapen.org/rest/search?query=publisher.name:ucl_press&expand=metadata"
                    httpretty.register_uri(httpretty.GET, url, body='[{"uuid":"df73bf94-b818-494c-a8dd-6775b0573bc2"}]')
                    # mock response of cloud function
                    mock_authorized_session.return_value = MagicMock(
                        spec=Response,
                        status_code=200,
                        json=lambda: {"entries": 100, "unprocessed_publishers": None},
                        reason="unit test",
                    )
                    url = "https://oapen-access-stats-kkinbzaigla-ew.a.run.app"
                    httpretty.register_uri(httpretty.POST, url, body="")
                    ti = env.run_task(telescope.call_cloud_function.__name__)
                    self.assertEqual(ti.state, State.SUCCESS)

                # Test transfer task
                gcs_upload_file(
                    bucket_name=IrusOapenTelescope.OAPEN_BUCKET,
                    blob_name=release.blob_name,
                    file_path=self.download_path,
                )
                ti = env.run_task(telescope.transfer.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_blob_integrity(env.download_bucket, release.blob_name, self.download_path)

                # Test download_transform task
                ti = env.run_task(telescope.download_transform.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assertTrue(os.path.exists(release.transform_path))
                self.assert_file_integrity(release.transform_path, "0b111b2f", "gzip_crc")

                # Test that transformed file uploaded
                ti = env.run_task(telescope.upload_transformed.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_blob_integrity(
                    env.transform_bucket, gcs_blob_name_from_path(release.transform_path), release.transform_path
                )

                # Test that data loaded into BigQuery
                ti = env.run_task(telescope.bq_load.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                table_id = bq_table_id(
                    project_id=telescope.cloud_workspace.project_id,
                    dataset_id=telescope.data_partner.bq_dataset_id,
                    table_id=telescope.data_partner.bq_table_name,
                )
                self.assert_table_integrity(table_id, 2)

                # Delete oapen bucket
                env._delete_bucket(IrusOapenTelescope.OAPEN_BUCKET)

                # Add_dataset_release_task
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 0)
                ti = env.run_task(telescope.add_new_dataset_releases.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                dataset_releases = get_dataset_releases(dag_id=telescope.dag_id, dataset_id=telescope.api_dataset_id)
                self.assertEqual(len(dataset_releases), 1)

                # Test that all telescope data deleted
                ti = env.run_task(telescope.cleanup.__name__)
                self.assertEqual(ti.state, State.SUCCESS)
                self.assert_cleanup(release.workflow_folder)

    @patch("observatory.platform.airflow.Variable.get")
    @patch("oaebu_workflows.irus_oapen_telescope.irus_oapen_telescope.upload_source_code_to_bucket")
    @patch("oaebu_workflows.irus_oapen_telescope.irus_oapen_telescope.cloud_function_exists")
    @patch("oaebu_workflows.irus_oapen_telescope.irus_oapen_telescope.create_cloud_function")
    def test_create_cloud_function(self, mock_create_function, mock_function_exists, mock_upload, mock_variable_get):
        """Test the create_cloud_function method of the IrusOapenRelease

        :param mock_variable_get: Mock Airflow Variable 'data'
        """

        def reset_mocks():
            mock_upload.reset_mock()
            mock_function_exists.reset_mock()
            mock_create_function.reset_mock()

        def assert_mocks(create: bool, update: bool):
            mock_upload.assert_called_once_with(
                telescope.FUNCTION_SOURCE_URL,
                telescope.OAPEN_PROJECT_ID,
                telescope.OAPEN_BUCKET,
                telescope.FUNCTION_BLOB_NAME,
                release.cloud_function_path,
            )
            mock_function_exists.assert_called_once_with(ANY, full_name)
            if create or update:
                mock_create_function.assert_called_once_with(
                    ANY,
                    location,
                    full_name,
                    telescope.OAPEN_BUCKET,
                    telescope.FUNCTION_BLOB_NAME,
                    telescope.max_active_runs,
                    update,
                )
            else:
                mock_create_function.assert_not_called()

        with CliRunner().isolated_filesystem():
            mock_variable_get.return_value = os.path.join(os.getcwd(), "data")
            # Create release instance
            cloud_workspace = CloudWorkspace(
                project_id="test-project",
                download_bucket="download_bucket",
                transform_bucket="transform_bucket",
                data_location="us",
            )
            telescope = IrusOapenTelescope(
                dag_id="irus_oapen_test",
                cloud_workspace=cloud_workspace,
                publisher_name_v4="publisher",
                publisher_uuid_v5="publisherUUID",
                bq_dataset_id="dataset_id",
            )
            release = IrusOapenRelease(
                dag_id=telescope.dag_id, run_id=random_id(), partition_date=pendulum.parse("2020-02-01")
            )
            location = f"projects/{telescope.OAPEN_PROJECT_ID}/locations/{telescope.FUNCTION_REGION}"
            full_name = f"{location}/functions/{telescope.FUNCTION_NAME}"

            # Test when source code upload was unsuccessful
            mock_upload.return_value = False, False
            task_instance = MagicMock()
            # context = dict(ti=task_instance)
            with self.assertRaises(AirflowException):
                telescope.create_cloud_function(releases=[release], ti=task_instance)

            # Test when cloud function does not exist
            reset_mocks()
            mock_upload.return_value = True, True
            mock_function_exists.return_value = False
            mock_create_function.return_value = True, "response"
            telescope.create_cloud_function(releases=[release], ti=task_instance)
            assert_mocks(create=True, update=False)

            # Test when cloud function exists, but source code has changed
            reset_mocks()
            mock_upload.return_value = True, True
            mock_function_exists.return_value = True
            mock_create_function.return_value = True, "response"
            telescope.create_cloud_function(telescope.max_active_runs)
            assert_mocks(create=False, update=True)

            # Test when cloud function exists and source code has not changed
            reset_mocks()
            mock_upload.return_value = True, False
            mock_function_exists.return_value = True
            telescope.create_cloud_function(releases=[release], ti=task_instance)
            assert_mocks(create=False, update=False)

            # Test when create cloud function was unsuccessful
            reset_mocks()
            mock_upload.return_value = True, True
            mock_function_exists.return_value = True
            mock_create_function.return_value = False, "response"
            with self.assertRaises(AirflowException):
                telescope.create_cloud_function(releases=[release], ti=task_instance)

    @patch("observatory.platform.airflow.Variable.get")
    @patch("oaebu_workflows.irus_oapen_telescope.irus_oapen_telescope.BaseHook.get_connection")
    @patch("oaebu_workflows.irus_oapen_telescope.irus_oapen_telescope.call_cloud_function")
    @patch("oaebu_workflows.irus_oapen_telescope.irus_oapen_telescope.cloud_function_exists")
    def test_call_cloud_function(self, mock_function_exists, mock_call_function, mock_conn_get, mock_variable_get):
        """Test the call_cloud_function method of the IrusOapenRelease

        :param mock_variable_get: Mock Airflow Variable 'data'
        """
        connections = {
            "geoip_license_key": Connection("geoip_license_key", uri="http://user_id:key@"),
            "irus_oapen_api": Connection("irus_oapen_api", uri="http://requestor_id:api_key@"),
            "irus_oapen_login": Connection("irus_oapen_login", uri="http://email:password@"),
        }
        mock_conn_get.side_effect = lambda x: connections[x]

        # Set URI to function url
        function_url = "https://oapen-access-stats-kkinbzfjal-ew.a.run.app"
        mock_function_exists.return_value = function_url

        with CliRunner().isolated_filesystem():
            mock_variable_get.return_value = os.path.join(os.getcwd(), "data")
            cloud_workspace = CloudWorkspace(
                project_id=self.project_id,
                download_bucket="download_bucket",
                transform_bucket="transform_bucket",
                data_location="us",
            )

            # Test new platform and old platform
            for date in ["2020-03", "2020-04"]:
                # Test for a given publisher name and the 'oapen' publisher
                for publisher in [("publisher", "uuid1"), ("oapen", "uuid2")]:
                    mock_call_function.reset_mock()

                    telescope = IrusOapenTelescope(
                        dag_id="irus_oapen_test",
                        cloud_workspace=cloud_workspace,
                        publisher_name_v4=publisher[0],
                        publisher_uuid_v5=publisher[1],
                        bq_dataset_id="dataset_id",
                    )
                    release = IrusOapenRelease(
                        dag_id=telescope.dag_id, run_id=random_id(), partition_date=pendulum.parse(date + "-01")
                    )
                    telescope.call_cloud_function(releases=[release])

                    # Test that the call function is called with the correct args
                    if date == "2020-04":
                        username = "requestor_id"
                        password = "api_key"
                    else:
                        username = "email"
                        password = "password"
                    mock_call_function.assert_called_once_with(
                        function_url,
                        date,
                        username,
                        password,
                        "key",
                        telescope.publisher_name_v4,
                        telescope.publisher_uuid_v5,
                        telescope.OAPEN_BUCKET,
                        release.blob_name,
                    )

    @patch("oaebu_workflows.irus_oapen_telescope.irus_oapen_telescope.gcs_upload_file")
    @patch("oaebu_workflows.irus_oapen_telescope.irus_oapen_telescope.gcs_create_bucket")
    def test_upload_source_code_to_bucket(self, mock_create_bucket, mock_upload_to_bucket):
        """Test getting source code from oapen irus uk release and uploading to storage bucket.
        Test expected results both when md5 hashes match and when they don't."""
        mock_create_bucket.return_value = True
        mock_upload_to_bucket.return_value = True, True
        with CliRunner().isolated_filesystem():
            cloud_function_path = os.path.join(os.getcwd(), "cloud_function.zip")
            success, upload = upload_source_code_to_bucket(
                IrusOapenTelescope.FUNCTION_SOURCE_URL,
                IrusOapenTelescope.OAPEN_PROJECT_ID,
                IrusOapenTelescope.OAPEN_BUCKET,
                IrusOapenTelescope.FUNCTION_BLOB_NAME,
                cloud_function_path,
            )
            self.assertEqual(success, True)
            self.assertEqual(upload, True)

            IrusOapenTelescope.FUNCTION_MD5_HASH = "different"
            with self.assertRaises(AirflowException):
                upload_source_code_to_bucket(
                    IrusOapenTelescope.FUNCTION_SOURCE_URL,
                    IrusOapenTelescope.OAPEN_PROJECT_ID,
                    IrusOapenTelescope.OAPEN_BUCKET,
                    IrusOapenTelescope.FUNCTION_BLOB_NAME,
                    cloud_function_path,
                )

    def test_cloud_function_exists(self):
        """Test the function that checks whether the cloud function exists"""
        requests = [
            # Cloud function exists
            {
                "response": (
                    None,
                    json.dumps(
                        {
                            "name": "projects/project-id/locations/us-central1/functions/function-2",
                            "serviceConfig": {"uri": "https://oapen-access-stats-kkinbzaigla-ew.a.run.app"},
                        }
                    ),
                ),
                "uri": "https://oapen-access-stats-kkinbzaigla-ew.a.run.app",
            },
            # Cloud function does not exist
            {"response": (httplib2.Response({"status": 404, "reason": "HttpError"}), b"{}"), "uri": None},
        ]
        for request in requests:
            with self.subTest(request=request):
                request_builder = RequestMockBuilder(
                    {
                        "cloudfunctions.projects.locations.functions.get": request["response"],
                    },
                    check_unexpected=True,
                )
                service = build(
                    "cloudfunctions",
                    "v2beta",
                    cache_discovery=False,
                    static_discovery=False,
                    requestBuilder=request_builder,
                )

                full_name = "projects/project-id/locations/us-central1/functions/function-2"
                uri = cloud_function_exists(service, full_name=full_name)
                self.assertEqual(request["uri"], uri)

    def test_create_cloud_function(self):
        """Test the function that creates the cloud function"""
        location = "projects/project-id/locations/us-central1"
        full_name = "projects/project-id/locations/us-central1/functions/function-2"
        source_bucket = "oapen_bucket"
        blob_name = "source_code.zip"
        max_active_runs = 1

        requests = [
            # Creating cloud function, no error
            {
                "method_id": "cloudfunctions.projects.locations.functions.create",
                "response": (
                    None,
                    json.dumps(
                        {
                            "name": "projects/project-id/locations/us-central1/operations/d29ya2Z",
                            "done": True,
                            "response": {"message": "response"},
                        }
                    ),
                ),
                "success": True,
                "msg": {"message": "response"},
            },
            # Updating/patching cloud function, no error
            {
                "method_id": "cloudfunctions.projects.locations.functions.patch",
                "response": (
                    None,
                    json.dumps(
                        {
                            "name": "projects/project-id/locations/us-central1/operations/d29ya2Z",
                            "done": True,
                            "response": {"message": "response"},
                        }
                    ),
                ),
                "success": True,
                "msg": {"message": "response"},
            },
            # Creating cloud function, error
            {
                "method_id": "cloudfunctions.projects.locations.functions.create",
                "response": (
                    None,
                    json.dumps(
                        {
                            "name": "projects/project-id/locations/us-central1/operations/d29ya2Z",
                            "done": True,
                            "error": {"message": "error"},
                        }
                    ),
                ),
                "success": False,
                "msg": {"message": "error"},
            },
        ]

        for request in requests:
            with self.subTest(request=request):
                request_builder = RequestMockBuilder(
                    {
                        request["method_id"]: (
                            None,
                            json.dumps(
                                {
                                    "name": "projects/project-id/locations/us-central1/operations/d29ya2Z",
                                    "done": False,
                                }
                            ),
                        ),
                        "cloudfunctions.projects.locations.operations.get": request["response"],
                    },
                    check_unexpected=True,
                )
                service = build(
                    "cloudfunctions",
                    "v2beta",
                    cache_discovery=False,
                    static_discovery=False,
                    requestBuilder=request_builder,
                )

                update = request["method_id"] == "cloudfunctions.projects.locations.functions.patch"
                success, msg = create_cloud_function(
                    service, location, full_name, source_bucket, blob_name, max_active_runs, update=update
                )
                self.assertEqual(request["success"], success)
                self.assertDictEqual(request["msg"], msg)

    @patch("oaebu_workflows.irus_oapen_telescope.irus_oapen_telescope.AuthorizedSession.post")
    def test_call_cloud_function(self, mock_authorized_session):
        """Test the function that calls the cloud function"""

        function_url = "function_url"
        release_date = "2020-01-01"
        username = "username"
        password = "password"
        geoip_license_key = "key"
        publisher_name = "publisher_name"
        publisher_uuid = "publisher_uuid"
        bucket_name = "bucket"
        blob_name = "blob"

        # Set responses for consequential calls
        mock_authorized_session.side_effect = [
            MagicMock(
                spec=Response,
                status_code=200,
                reason="unit test",
                json=lambda: {"entries": 100, "unprocessed_publishers": ["publisher1", "publisher2"]},
            ),
            MagicMock(
                spec=Response,
                status_code=200,
                reason="unit test",
                json=lambda: {"entries": 200, "unprocessed_publishers": None},
            ),
            MagicMock(
                spec=Response,
                status_code=200,
                reason="unit test",
                json=lambda: {"entries": 0, "unprocessed_publishers": None},
            ),
            MagicMock(spec=Response, status_code=400, reason="unit test"),
        ]
        # Test when there are unprocessed publishers (first 2 responses from side effect)
        call_cloud_function(
            function_url,
            release_date,
            username,
            password,
            geoip_license_key,
            publisher_name,
            publisher_uuid,
            bucket_name,
            blob_name,
        )
        self.assertEqual(2, mock_authorized_session.call_count)

        # Test when entries is 0 (3rd response from side effect)
        with self.assertRaises(AirflowSkipException):
            call_cloud_function(
                function_url,
                release_date,
                username,
                password,
                geoip_license_key,
                publisher_name,
                publisher_uuid,
                bucket_name,
                blob_name,
            )

        # Test when response status code is not 200 (last response from side effect)
        with self.assertRaises(AirflowException):
            call_cloud_function(
                function_url,
                release_date,
                username,
                password,
                geoip_license_key,
                publisher_name,
                publisher_uuid,
                bucket_name,
                blob_name,
            )
